# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Buildd Slave tests.

This file contains the follwoing tests:

 * Basic authentication handling (used to download private sources);
 * Build log sanitization (removal of passwords from private buildlog);
 * Build log(tail) mechanisms (limited output from the end of the buildlog).

"""

__metaclass__ = type

__all__ = ['LaunchpadBuilddSlaveTests']

import difflib
import os
import shutil
import tempfile
import urllib2
import unittest
import xmlrpclib

from lpbuildd.tests.harness import (
    BuilddSlaveTestSetup,
    BuilddTestCase,
    MockBuildManager,
    )


def read_file(path):
    """Helper for reading the contents of a file."""
    file_object = open(path)
    try:
        return file_object.read()
    finally:
        file_object.close()


class LaunchpadBuilddSlaveTests(BuilddTestCase):
    """Unit tests for scrubbing (removal of passwords) of buildlog files."""

    def testBasicAuth(self):
        """Test that the auth handler is installed with the right details."""
        url = "http://fakeurl/"
        user = "myuser"
        password = "fakepassword"

        opener = self.slave.setupAuthHandler(url, user, password)

        # Inspect the openers and ensure the wanted handler is installed.
        basic_auth_handler = None
        for handler in opener.handlers:
            if isinstance(handler, urllib2.HTTPBasicAuthHandler):
                basic_auth_handler = handler
                break
        self.assertTrue(
            basic_auth_handler is not None,
            "No basic auth handler installed.")

        password_mgr = basic_auth_handler.passwd
        stored_user, stored_pass = password_mgr.find_user_password(None, url)
        self.assertEqual(user, stored_user)
        self.assertEqual(password, stored_pass)

    def testBuildlogScrubbing(self):
        """Tests the buildlog scrubbing (removal of passwords from URLs)."""
        # This is where the buildlog file lives.
        log_path = self.slave.cachePath('buildlog')

        # This is where the slave leaves the original/unsanitized
        # buildlog file after scrubbing.
        unsanitized_path = self.slave.cachePath('buildlog.unsanitized')

        # Copy the fake buildlog file to the cache path.
        shutil.copy(os.path.join(self.here, 'buildlog'), log_path)

        # Invoke the slave's buildlog scrubbing method.
        self.slave.sanitizeBuildlog(log_path)

        # Read the unsanitized original content.
        unsanitized = read_file(unsanitized_path).splitlines()
        # Read the new, sanitized content.
        clean = read_file(log_path).splitlines()

        # Compare the scrubbed content with the unsanitized one.
        differences = '\n'.join(difflib.unified_diff(unsanitized, clean))

        # Read the expected differences from the prepared disk file.
        expected = read_file(os.path.join(self.here, 'test_1.diff'))

        # Python difflib slightly changed the formatting of the headers; we
        # don't care; let's make it look the same as it used to.
        differences = differences.replace(
            '--- \n\n+++ \n\n',
            '---  \n\n+++  \n\n')
        
        # Make sure they match.
        self.assertEqual(differences, expected)

    def testLogtailScrubbing(self):
        """Test the scrubbing of the slave's getLogTail() output."""

        # This is where the buildlog file lives.
        log_path = self.slave.cachePath('buildlog')

        # Copy the prepared, longer buildlog file so we can test lines
        # that are chopped off in the middle.
        shutil.copy(os.path.join(self.here, 'buildlog.long'), log_path)

        # First get the unfiltered log tail output (which is the default
        # behaviour because the BuildManager's 'is_archive_private'
        # property is initialized to False).
        self.slave.manager.is_archive_private = False
        unsanitized = self.slave.getLogTail().splitlines()

        # Make the slave believe we are building in a private archive to
        # obtain the scrubbed log tail output.
        self.slave.manager.is_archive_private = True
        clean = self.slave.getLogTail().splitlines()

        # Get the differences ..
        differences = '\n'.join(difflib.unified_diff(unsanitized, clean))

        # .. and the expected differences.
        expected = read_file(os.path.join(self.here, 'test_2.diff'))

        # Python difflib slightly changed the formatting of the headers; we
        # don't care; let's make it look the same as it used to.
        differences = differences.replace(
            '--- \n\n+++ \n\n',
            '---  \n\n+++  \n\n')
        
        # Finally make sure what we got is what we expected.
        self.assertEqual(differences, expected)

    def testLogtail(self):
        """Tests the logtail mechanisms.

        'getLogTail' return up to 2 KiB text from the current 'buildlog' file.
        """
        self.makeLog(0)
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 0)

        self.makeLog(1)
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 1)

        self.makeLog(2048)
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 2048)

        self.makeLog(2049)
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 2048)

        self.makeLog(4096)
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 2048)

    def testLogtailWhenLogFileVanishes(self):
        """Slave.getLogTail doesn't get hurt if the logfile has vanished.

        This is a common race-condition in our slaves, since they get
        pooled all the time when they are building.

        Sometimes the getLogTail calls coincides with the job
        cleanup/sanitization, so there is no buildlog to inspect and thus
        we expect an empty string to be returned instead of a explosion.
        """
        # Create some log content and read it.
        self.makeLog(2048)
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 2048)

        # Read it again for luck.
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 2048)

        # Remove the buildlog file
        os.remove(self.slave.cachePath('buildlog'))

        # Instead of shocking the getLogTail call, return an empty string.
        log_tail = self.slave.getLogTail()
        self.assertEqual(len(log_tail), 0)

    def testCleanDuplicateFiles(self):
        """The clean method copes with duplicate waiting files."""
        workdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, workdir)
        self.slave._log = None
        self.slave.startBuild(MockBuildManager())
        self.slave.buildComplete()
        paths = [os.path.join(workdir, name) for name in ('a', 'b')]
        for path in paths:
            f = open(path, 'w')
            f.write('data')
            f.close()
            self.slave.addWaitingFile(path)
        self.slave.clean()
        self.assertEqual([], os.listdir(self.slave._cachepath))


class XMLRPCBuildDSlaveTests(unittest.TestCase):

    def setUp(self):
        super(XMLRPCBuildDSlaveTests, self).setUp()
        self.slave = BuilddSlaveTestSetup()
        self.slave.setUp()
        self.server = xmlrpclib.Server('http://localhost:8221/rpc/')

    def tearDown(self):
        self.slave.tearDown()
        super(XMLRPCBuildDSlaveTests, self).tearDown()

    def test_build_unknown_builder(self):
        # If a bogus builder name is passed into build, it returns an
        # appropriate error message and not just 'None'.
        buildername = 'nonexistentbuilder'
        status, info = self.server.build('foo', buildername, 'sha1', {}, {})

        self.assertEqual('BuilderStatus.UNKNOWNBUILDER', status)
        self.assertTrue(
            info is not None, "UNKNOWNBUILDER returns 'None' info.")
        self.assertTrue(
            info.startswith("%s not in [" % buildername),
            'UNKNOWNBUILDER info is "%s"' % info)



#! /usr/bin/python
# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os.path
import sys
import tarfile

import logging

from bzrlib.branch import Branch
from bzrlib.export import export

from lpbuildd.pottery import intltool


class GenerateTranslationTemplates:
    """Script to generate translation templates from a branch."""

    def __init__(self, branch_spec, result_name, work_dir, log_file=None):
        """Prepare to generate templates for a branch.

        :param branch_spec: Either a branch URL or the path of a local
            branch.  URLs are recognized by the occurrence of ':'.  In
            the case of a URL, this will make up a path for the branch
            and check out the branch to there.
        :param result_name: The name of the result tarball. Should end in
            .tar.gz.
        :param work_dir: The directory to work in. Must exist.
        :param log_file: File-like object to log to. If None, defaults to
            stderr.
        """
        self.work_dir = work_dir
        self.branch_spec = branch_spec
        self.result_name = result_name
        self.logger = self._setupLogger(log_file)

    def _setupLogger(self, log_file):
        """Sets up and returns a logger."""
        if log_file is None:
            log_file = sys.stderr
        logger = logging.getLogger("generate-templates")
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(log_file)
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)
        return logger

    def _getBranch(self):
        """Set `self.branch_dir`, and check out branch if needed."""
        if ':' in self.branch_spec:
            # This is a branch URL.  Check out the branch.
            self.branch_dir = os.path.join(self.work_dir, 'source-tree')
            self.logger.info("Getting remote branch %s..." % self.branch_spec)
            self._checkout(self.branch_spec)
        else:
            # This is a local filesystem path.  Use the branch in-place.
            self.logger.info("Using local branch %s..." % self.branch_spec)
            self.branch_dir = self.branch_spec

    def _checkout(self, branch_url):
        """Check out a source branch to generate from.

        The branch is checked out to the location specified by
        `self.branch_dir`.
        """
        self.logger.info("Opening branch %s..." % branch_url)
        branch = Branch.open(branch_url)
        self.logger.info("Getting branch revision tree...")
        rev_tree = branch.basis_tree()
        self.logger.info("Exporting branch to %s..." % self.branch_dir)
        export(rev_tree, self.branch_dir)
        self.logger.info("Exporting branch done.")

    def _makeTarball(self, files):
        """Put the given files into a tarball in the working directory."""
        tarname = os.path.join(self.work_dir, self.result_name)
        self.logger.info("Making tarball with templates in %s..." % tarname)
        tarball = tarfile.open(tarname, 'w|gz')
        files = [name for name in files if not name.endswith('/')]
        for path in files:
            full_path = os.path.join(self.branch_dir, path)
            self.logger.info("Adding template %s..." % full_path)
            tarball.add(full_path, path)
        tarball.close()
        self.logger.info("Tarball generated.")

    def generate(self):
        """Do It.  Generate templates."""
        self.logger.info("Generating templates for %s." % self.branch_spec)
        self._getBranch()
        pots = intltool.generate_pots(self.branch_dir)
        self.logger.info("Generated %d templates." % len(pots))
        if len(pots) > 0:
            self._makeTarball(pots)
        return 0


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "Usage: %s branch resultname [workdir]" % sys.argv[0]
        print "  'branch' is a branch URL or directory."
        print "  'resultname' is the name of the result tarball."
        print "  'workdir' is a directory, defaults to HOME."
        sys.exit(1)
    if len(sys.argv) == 4:
        workdir = sys.argv[3]
    else:
        workdir = os.environ['HOME']
    script = GenerateTranslationTemplates(
        sys.argv[1], sys.argv[2], workdir)
    sys.exit(script.generate())



# Copyright 2009, 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import absolute_import

from collections import defaultdict
import os
import re
import subprocess
import traceback

import apt_pkg
from debian.deb822 import (
    Dsc,
    PkgRelation,
    )
from debian.debian_support import Version

from lpbuildd.debian import (
    DebianBuildManager,
    DebianBuildState,
    )


class SBuildExitCodes:
    """SBUILD process result codes."""
    OK = 0
    FAILED = 1
    ATTEMPTED = 2
    GIVENBACK = 3
    BUILDERFAIL = 4


APT_MISSING_DEP_PATTERNS = [
    'but [^ ]* is to be installed',
    'but [^ ]* is installed',
    'but it is not installable',
    'but it is a virtual package',
    ]


APT_DUBIOUS_DEP_PATTERNS = [
    'but it is not installed',
    'but it is not going to be installed',
    ]


class BuildLogRegexes:
    """Build log regexes for performing actions based on regexes, and extracting dependencies for auto dep-waits"""
    GIVENBACK = [
        ("^E: There are problems and -y was used without --force-yes"),
        ]
    MAYBEDEPFAIL = [
        'The following packages have unmet dependencies:\n'
        '.* Depends: [^ ]*( \([^)]*\))? (%s)\n'
        % '|'.join(APT_DUBIOUS_DEP_PATTERNS),
        ]
    DEPFAIL = {
        'The following packages have unmet dependencies:\n'
        '.* Depends: (?P<p>[^ ]*( \([^)]*\))?) (%s)\n'
        % '|'.join(APT_MISSING_DEP_PATTERNS): "\g<p>",
        }


class DpkgArchitectureCache:
    """Cache the results of asking questions of dpkg-architecture."""

    def __init__(self):
        self._matches = {}

    def match(self, arch, wildcard):
        if (arch, wildcard) not in self._matches:
            command = ["dpkg-architecture", "-i%s" % wildcard]
            env = dict(os.environ)
            env["DEB_HOST_ARCH"] = arch
            ret = (subprocess.call(command, env=env) == 0)
            self._matches[(arch, wildcard)] = ret
        return self._matches[(arch, wildcard)]


dpkg_architecture = DpkgArchitectureCache()


class BinaryPackageBuildState(DebianBuildState):
    SBUILD = "SBUILD"


class BinaryPackageBuildManager(DebianBuildManager):
    """Handle buildd building for a debian style binary package build"""

    initial_build_state = BinaryPackageBuildState.SBUILD

    def __init__(self, slave, buildid, **kwargs):
        DebianBuildManager.__init__(self, slave, buildid, **kwargs)
        self._sbuildpath = os.path.join(self._slavebin, "sbuild-package")

    @property
    def chroot_path(self):
        return os.path.join(
            self.home, "build-" + self._buildid, 'chroot-autobuild')

    def initiate(self, files, chroot, extra_args):
        """Initiate a build with a given set of files and chroot."""

        self._dscfile = None
        for f in files:
            if f.endswith(".dsc"):
                self._dscfile = f
        if self._dscfile is None:
            raise ValueError, files

        self.archive_purpose = extra_args.get('archive_purpose')
        self.distribution = extra_args['distribution']
        self.suite = extra_args['suite']
        self.component = extra_args['ogrecomponent']
        self.arch_indep = extra_args.get('arch_indep', False)
        self.build_debug_symbols = extra_args.get('build_debug_symbols', False)

        super(BinaryPackageBuildManager, self).initiate(
            files, chroot, extra_args)

    def doRunBuild(self):
        """Run the sbuild process to build the package."""
        currently_building_path = os.path.join(
            self.chroot_path, 'CurrentlyBuilding')
        currently_building_contents = (
            'Package: %s\n'
            'Component: %s\n'
            'Suite: %s\n'
            'Purpose: %s\n'
            % (self._dscfile.split('_')[0], self.component, self.suite,
               self.archive_purpose))
        if self.build_debug_symbols:
            currently_building_contents += 'Build-Debug-Symbols: yes\n'
        with open(currently_building_path, 'w') as currently_building:
            currently_building.write(currently_building_contents)

        args = ["sbuild-package", self._buildid, self.arch_tag]
        args.append(self.suite)
        args.extend(["-c", "chroot:autobuild"])
        args.append("--arch=" + self.arch_tag)
        args.append("--dist=" + self.suite)
        args.append("--purge=never")
        args.append("--nolog")
        if self.arch_indep:
            args.append("-A")
        args.append(self._dscfile)
        self.runSubProcess(self._sbuildpath, args)

    def getAvailablePackages(self):
        """Return the available binary packages in the chroot.

        :return: A dictionary mapping package names to a set of the
            available versions of each package.
        """
        available = defaultdict(set)
        apt_lists = os.path.join(
            self.chroot_path, "var", "lib", "apt", "lists")
        for name in sorted(os.listdir(apt_lists)):
            if name.endswith("_Packages"):
                path = os.path.join(apt_lists, name)
                with open(path, "rb") as packages_file:
                    for section in apt_pkg.TagFile(packages_file):
                        available[section["package"]].add(section["version"])
                        if "provides" in section:
                            provides = apt_pkg.parse_depends(
                                section["provides"])
                            for provide in provides:
                                # Disjunctions are currently undefined here.
                                if len(provide) > 1:
                                    continue
                                # Virtual packages may only provide an exact
                                # version or none.
                                if provide[0][1] and provide[0][2] != "=":
                                    continue
                                available[provide[0][0]].add(
                                    provide[0][1] if provide[0][1] else None)
        return available

    def getBuildDepends(self, dscpath, arch_indep):
        """Get the build-dependencies of a source package.

        :param dscpath: The path of a .dsc file.
        :param arch_indep: True iff we were asked to build the
            architecture-independent parts of this source package.
        :return: The build-dependencies, in the form returned by
            `debian.deb822.PkgRelation.parse_relations`.
        """
        deps = []
        with open(dscpath, "rb") as dscfile:
            dsc = Dsc(dscfile)
            fields = ["Build-Depends"]
            if arch_indep:
                fields.append("Build-Depends-Indep")
            for field in fields:
                if field in dsc:
                    deps.extend(PkgRelation.parse_relations(dsc[field]))
        return deps

    def relationMatches(self, dep, available):
        """Return True iff a dependency matches an available package.

        :param dep: A dictionary with at least a "name" key, perhaps also
            "version", "arch", and "restrictions" keys, and optionally other
            keys, of the kind returned in a list of lists by
            `debian.deb822.PkgRelation.parse_relations`.
        :param available: A dictionary mapping package names to a list of
            the available versions of each package.
        """
        dep_arch = dep.get("arch")
        if dep_arch is not None:
            arch_match = False
            for enabled, arch_wildcard in dep_arch:
                if dpkg_architecture.match(self.arch_tag, arch_wildcard):
                    arch_match = enabled
                    break
                elif not enabled:
                    # Any !other-architecture restriction implies that this
                    # architecture is allowed, unless it's specifically
                    # excluded by some other restriction.
                    arch_match = True
            if not arch_match:
                # This dependency "matches" in the sense that it's ignored
                # on this architecture.
                return True
        dep_restrictions = dep.get("restrictions")
        if dep_restrictions is not None:
            if all(
                any(restriction.enabled for restriction in restrlist)
                for restrlist in dep_restrictions):
                # This dependency "matches" in the sense that it's ignored
                # when no build profiles are enabled.
                return True
        if dep["name"] not in available:
            return False
        dep_version = dep.get("version")
        if dep_version is None:
            return True
        operator_map = {
            "<<": (lambda a, b: a < b),
            "<=": (lambda a, b: a <= b),
            "=": (lambda a, b: a == b),
            ">=": (lambda a, b: a >= b),
            ">>": (lambda a, b: a > b),
            }
        operator = operator_map[dep_version[0]]
        want_version = dep_version[1]
        for version in available[dep["name"]]:
            if (version is not None and
                    operator(Version(version), want_version)):
                return True
        return False

    def stripDependencies(self, deps):
        """Return a stripped and stringified representation of a dependency.

        The build master can't handle the various qualifications and
        restrictions that may be present in control-format
        build-dependencies (e.g. ":any", "[amd64]", or "<!nocheck>"), so we
        strip these out before returning them.

        :param deps: Build-dependencies in the form returned by
            `debian.deb822.PkgRelation.parse_relations`.
        :return: A stripped dependency relation string, or None if deps is
            empty.
        """
        stripped_deps = []
        for or_dep in deps:
            stripped_or_dep = []
            for simple_dep in or_dep:
                stripped_simple_dep = dict(simple_dep)
                stripped_simple_dep["arch"] = None
                stripped_simple_dep["archqual"] = None
                stripped_simple_dep["restrictions"] = None
                stripped_or_dep.append(stripped_simple_dep)
            stripped_deps.append(stripped_or_dep)
        if stripped_deps:
            return PkgRelation.str(stripped_deps)
        else:
            return None

    def analyseDepWait(self, deps, avail):
        """Work out the correct dep-wait for a failed build, if any.

        We only consider direct build-dependencies, because indirect ones
        can't easily be turned into an accurate dep-wait: they might be
        resolved either by an intermediate package changing or by the
        missing dependency becoming available.  We err on the side of
        failing a build rather than setting a dep-wait if it's possible that
        the dep-wait might be incorrect.  Any exception raised during the
        analysis causes the build to be failed.

        :param deps: The source package's build-dependencies, in the form
            returned by `debian.deb822.PkgRelation.parse_relations`.
        :param avail: A dictionary mapping package names to a set of the
            available versions of each package.
        :return: A dependency relation string representing the packages that
            need to become available before this build can proceed, or None
            if the build should be failed instead.
        """
        try:
            unsat_deps = []
            for or_dep in deps:
                if not any(self.relationMatches(dep, avail) for dep in or_dep):
                    unsat_deps.append(or_dep)
            return self.stripDependencies(unsat_deps)
        except Exception:
            self._slave.log("Failed to analyse dep-wait:\n")
            for line in traceback.format_exc().splitlines(True):
                self._slave.log(line)
            return None

    def iterate_SBUILD(self, success):
        """Finished the sbuild run."""
        if success == SBuildExitCodes.OK:
            print("Returning build status: OK")
            try:
                self.gatherResults()
            except Exception, e:
                self._slave.log("Failed to gather results: %s" % e)
                self._slave.buildFail()
                self.alreadyfailed = True
            self.doReapProcesses(self._state)
            return

        log_patterns = []
        stop_patterns = [["^Toolchain package versions:", re.M]]

        # We don't distinguish attempted and failed.
        if success == SBuildExitCodes.ATTEMPTED:
            success = SBuildExitCodes.FAILED

        if success == SBuildExitCodes.GIVENBACK:
            for rx in BuildLogRegexes.GIVENBACK:
                log_patterns.append([rx, re.M])
            # Check the last 4KiB for the Fail-Stage. If it failed
            # during install-deps, search for the missing dependency
            # string.
            with open(os.path.join(self._cachepath, "buildlog")) as log:
                try:
                    log.seek(-4096, os.SEEK_END)
                except IOError:
                    pass
                tail = log.read(4096)
            if re.search("^Fail-Stage: install-deps$", tail, re.M):
                for rx in BuildLogRegexes.MAYBEDEPFAIL:
                    log_patterns.append([rx, re.M | re.S])
                for rx in BuildLogRegexes.DEPFAIL:
                    log_patterns.append([rx, re.M | re.S])

        missing_dep = None
        if log_patterns:
            rx, mo = self.searchLogContents(log_patterns, stop_patterns)
            if mo is None:
                # It was givenback, but we can't see a valid reason.
                # Assume it failed.
                success = SBuildExitCodes.FAILED
            elif rx in BuildLogRegexes.MAYBEDEPFAIL:
                # These matches need further analysis.
                dscpath = os.path.join(
                    self.home, "build-%s" % self._buildid, self._dscfile)
                missing_dep = self.analyseDepWait(
                    self.getBuildDepends(dscpath, self.arch_indep),
                    self.getAvailablePackages())
                if missing_dep is None:
                    success = SBuildExitCodes.FAILED
            elif rx in BuildLogRegexes.DEPFAIL:
                # A depwait match forces depwait.
                missing_dep = mo.expand(BuildLogRegexes.DEPFAIL[rx])
                missing_dep = self.stripDependencies(
                    PkgRelation.parse_relations(missing_dep))
            else:
                # Otherwise it was a givenback pattern, so leave it
                # in givenback.
                pass

        if not self.alreadyfailed:
            if missing_dep is not None:
                print("Returning build status: DEPFAIL")
                print("Dependencies: " + missing_dep)
                self._slave.depFail(missing_dep)
            elif success == SBuildExitCodes.GIVENBACK:
                print("Returning build status: GIVENBACK")
                self._slave.giveBack()
            elif success == SBuildExitCodes.FAILED:
                print("Returning build status: PACKAGEFAIL")
                self._slave.buildFail()
            elif success >= SBuildExitCodes.BUILDERFAIL:
                # anything else is assumed to be a buildd failure
                print("Returning build status: BUILDERFAIL")
                self._slave.builderFail()
            self.alreadyfailed = True
        self.doReapProcesses(self._state)

    def iterateReap_SBUILD(self, success):
        """Finished reaping after sbuild run."""
        self._state = DebianBuildState.UMOUNT
        self.doUnmounting()



# Copyright 2009, 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

# Authors: Daniel Silverstone <daniel.silverstone@canonical.com>
#      and Adam Conrad <adam.conrad@canonical.com>

# Buildd Slave implementation

__metaclass__ = type

from functools import partial
import hashlib
import os
import re
import urllib2
import warnings
import xmlrpclib

import apt
from twisted.internet import protocol
from twisted.internet import reactor as default_reactor
from twisted.internet import process
from twisted.python import log
from twisted.web import xmlrpc

# XXX cjwatson 2013-10-07: Remove when all builders are on Ubuntu 10.04 LTS
# or newer.
warnings.filterwarnings(
    "ignore", "apt API not stable yet", category=FutureWarning)

devnull = open("/dev/null", "r")


def _sanitizeURLs(text_seq):
    """A generator that deletes URL passwords from a string sequence.

    This generator removes user/password data from URLs if embedded
    in the latter as follows: scheme://user:passwd@netloc/path.

    :param text_seq: A sequence of strings (that may contain URLs).
    :return: A (sanitized) line stripped of authentication credentials.
    """
    # This regular expression will be used to remove authentication
    # credentials from URLs.
    password_re = re.compile('://([^:]+:[^@]+@)(\S+)')

    for line in text_seq:
        sanitized_line = password_re.sub(r'://\2', line)
        yield sanitized_line


# XXX cprov 2005-06-28:
# RunCapture can be replaced with a call to
#
#   twisted.internet.utils.getProcessOutputAndValue
#
# when we start using Twisted 2.0.
class RunCapture(protocol.ProcessProtocol):
    """Run a command and capture its output to a slave's log"""

    def __init__(self, slave, callback):
        self.slave = slave
        self.notify = callback
        self.builderFailCall = None
        self.ignore = False

    def outReceived(self, data):
        """Pass on stdout data to the log."""
        self.slave.log(data)

    def errReceived(self, data):
        """Pass on stderr data to the log.

        With a bit of luck we won't interleave horribly."""
        self.slave.log(data)

    def processEnded(self, statusobject):
        """This method is called when a child process got terminated.

        Two actions are required at this point: eliminate pending calls to
        "builderFail", and invoke the programmed notification callback.  The
        notification callback must be invoked last.
        """
        if self.ignore:
            # The build manager no longer cares about this process.
            return

        # Since the process terminated, we don't need to fail the builder.
        if self.builderFailCall and self.builderFailCall.active():
            self.builderFailCall.cancel()

        # notify the slave, it'll perform the required actions
        if self.notify is not None:
            self.notify(statusobject.value.exitCode)


class BuildManager(object):
    """Build Daemon slave build manager abstract parent"""

    def __init__(self, slave, buildid, reactor=None):
        """Create a BuildManager.

        :param slave: A `BuildDSlave`.
        :param buildid: Identifying string for this build.
        """
        object.__init__(self)
        self._buildid = buildid
        self._slave = slave
        if reactor is None:
            reactor = default_reactor
        self._reactor = reactor
        self._sharepath = slave._config.get("slave", "sharepath")
        self._slavebin = os.path.join(self._sharepath, "slavebin")
        self._preppath = os.path.join(self._slavebin, "slave-prep")
        self._unpackpath = os.path.join(self._slavebin, "unpack-chroot")
        self._cleanpath = os.path.join(self._slavebin, "remove-build")
        self._mountpath = os.path.join(self._slavebin, "mount-chroot")
        self._umountpath = os.path.join(self._slavebin, "umount-chroot")
        self._scanpath = os.path.join(self._slavebin, "scan-for-processes")
        self._subprocess = None
        self._reaped_states = set()
        self.is_archive_private = False
        self.home = os.environ['HOME']
        self.abort_timeout = 120

    def runSubProcess(self, command, args, iterate=None):
        """Run a sub process capturing the results in the log."""
        if iterate is None:
            iterate = self.iterate
        self._subprocess = RunCapture(self._slave, iterate)
        self._slave.log("RUN: %s %r\n" % (command, args))
        childfds = {0: devnull.fileno(), 1: "r", 2: "r"}
        self._reactor.spawnProcess(
            self._subprocess, command, args, env=os.environ,
            path=self.home, childFDs=childfds)

    def doUnpack(self):
        """Unpack the build chroot."""
        self.runSubProcess(
            self._unpackpath,
            ["unpack-chroot", self._buildid, self._chroottarfile])

    def doReapProcesses(self, state, notify=True):
        """Reap any processes left lying around in the chroot."""
        if state is not None and state in self._reaped_states:
            # We've already reaped this state.  To avoid a loop, proceed
            # immediately to the next iterator.
            self._slave.log("Already reaped from state %s...\n" % state)
            if notify:
                self.iterateReap(state, 0)
        else:
            if state is not None:
                self._reaped_states.add(state)
            if notify:
                iterate = partial(self.iterateReap, state)
            else:
                iterate = lambda success: None
            self.runSubProcess(
                self._scanpath, ["scan-for-processes", self._buildid],
                iterate=iterate)

    def doCleanup(self):
        """Remove the build tree etc."""
        self.runSubProcess(self._cleanpath, ["remove-build", self._buildid])

        # Sanitize the URLs in the buildlog file if this is a build
        # in a private archive.
        if self.is_archive_private:
            self._slave.sanitizeBuildlog(self._slave.cachePath("buildlog"))

    def doMounting(self):
        """Mount things in the chroot, e.g. proc."""
        self.runSubProcess( self._mountpath,
                            ["mount-chroot", self._buildid])

    def doUnmounting(self):
        """Unmount the chroot."""
        self.runSubProcess( self._umountpath,
                            ["umount-chroot", self._buildid])

    def initiate(self, files, chroot, extra_args):
        """Initiate a build given the input files.

        Please note: the 'extra_args' dictionary may contain a boolean
        value keyed under the 'archive_private' string. If that value
        evaluates to True the build at hand is for a private archive.
        """
        os.mkdir("%s/build-%s" % (self.home, self._buildid))
        for f in files:
            os.symlink( self._slave.cachePath(files[f]),
                        "%s/build-%s/%s" % (self.home,
                                            self._buildid, f))
        self._chroottarfile = self._slave.cachePath(chroot)

        # Check whether this is a build in a private archive and
        # whether the URLs in the buildlog file should be sanitized
        # so that they do not contain any embedded authentication
        # credentials.
        if extra_args.get('archive_private'):
            self.is_archive_private = True

        self.runSubProcess(self._preppath, ["slave-prep"])

    def iterate(self, success):
        """Perform an iteration of the slave.

        The BuildManager tends to work by invoking several
        subprocesses in order. the iterate method is called by the
        object created by runSubProcess to gather the results of the
        sub process.
        """
        raise NotImplementedError("BuildManager should be subclassed to be "
                                  "used")

    def iterateReap(self, state, success):
        """Perform an iteration of the slave following subprocess reaping.

        Subprocess reaping is special, typically occurring at several
        positions in a build manager's state machine.  We therefore keep
        track of the state being reaped so that we can select the
        appropriate next state.
        """
        raise NotImplementedError("BuildManager should be subclassed to be "
                                  "used")

    def abortReap(self):
        """Abort by killing all processes in the chroot, as hard as we can.

        We expect this to result in the main build process exiting non-zero
        and giving us some useful logs.

        This may be overridden in subclasses so that they can perform their
        own state machine management.
        """
        self.doReapProcesses(None, notify=False)

    def abort(self):
        """Abort the build by killing the subprocess."""
        if self.alreadyfailed or self._subprocess is None:
            return
        else:
            self.alreadyfailed = True
        primary_subprocess = self._subprocess
        self.abortReap()
        # In extreme cases the build may be hung too badly for
        # scan-for-processes to manage to kill it (blocked on I/O,
        # forkbombing test suite, etc.).  In this case, fail the builder and
        # let an admin sort it out.
        self._subprocess.builderFailCall = self._reactor.callLater(
            self.abort_timeout, self.builderFail,
            "Failed to kill all processes.", primary_subprocess)

    def builderFail(self, reason, primary_subprocess):
        """Mark the builder as failed."""
        self._slave.log("ABORTING: %s\n" % reason)
        self._subprocess.builderFailCall = None
        self._slave.builderFail()
        self.alreadyfailed = True
        # If we failed to kill all processes in the chroot, then the primary
        # subprocess (i.e. the one running immediately before
        # doReapProcesses was called) may not have exited.  Kill it so that
        # we can proceed.
        try:
            primary_subprocess.transport.signalProcess('KILL')
        except process.ProcessExitedAlready:
            self._slave.log("ABORTING: Process Exited Already\n")
        primary_subprocess.transport.loseConnection()
        # Leave the reaper running, but disconnect it from our state
        # machine.  Perhaps an admin can make something of it, and in any
        # case scan-for-processes elevates itself to root so it's awkward to
        # kill it.
        self._subprocess.ignore = True
        self._subprocess.transport.loseConnection()


class BuilderStatus:
    """Status values for the builder."""

    IDLE = "BuilderStatus.IDLE"
    BUILDING = "BuilderStatus.BUILDING"
    WAITING = "BuilderStatus.WAITING"
    ABORTING = "BuilderStatus.ABORTING"

    UNKNOWNSUM = "BuilderStatus.UNKNOWNSUM"
    UNKNOWNBUILDER = "BuilderStatus.UNKNOWNBUILDER"


class BuildStatus:
    """Status values for builds themselves."""

    OK = "BuildStatus.OK"
    DEPFAIL = "BuildStatus.DEPFAIL"
    GIVENBACK = "BuildStatus.GIVENBACK"
    PACKAGEFAIL = "BuildStatus.PACKAGEFAIL"
    CHROOTFAIL = "BuildStatus.CHROOTFAIL"
    BUILDERFAIL = "BuildStatus.BUILDERFAIL"
    ABORTED = "BuildStatus.ABORTED"


class BuildDSlave(object):
    """Build Daemon slave. Implementation of most needed functions
    for a Build-Slave device.
    """

    def __init__(self, config):
        object.__init__(self)
        self._config = config
        self.builderstatus = BuilderStatus.IDLE
        self._cachepath = self._config.get("slave","filecache")
        self.buildstatus = BuildStatus.OK
        self.waitingfiles = {}
        self.builddependencies = ""
        self._log = None

        if not os.path.isdir(self._cachepath):
            raise ValueError("FileCache path is not a dir")

    def getArch(self):
        """Return the Architecture tag for the slave."""
        return self._config.get("slave","architecturetag")

    def cachePath(self, file):
        """Return the path in the cache of the file specified."""
        return os.path.join(self._cachepath, file)

    def setupAuthHandler(self, url, username, password):
        """Set up a BasicAuthHandler to open the url.

        :param url: The URL that needs authenticating.
        :param username: The username for authentication.
        :param password: The password for authentication.
        :return: The OpenerDirector instance.

        This helper installs a urllib2.HTTPBasicAuthHandler that will deal
        with any HTTP basic authentication required when opening the
        URL.
        """
        password_mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
        password_mgr.add_password(None, url, username, password)
        handler = urllib2.HTTPBasicAuthHandler(password_mgr)
        opener = urllib2.build_opener(handler)
        return opener

    def ensurePresent(self, sha1sum, url=None, username=None, password=None):
        """Ensure we have the file with the checksum specified.

        Optionally you can provide the librarian URL and
        the build slave will fetch the file if it doesn't have it.
        Return a tuple containing: (<present>, <info>)
        """
        extra_info = 'No URL'
        cachefile = self.cachePath(sha1sum)
        if url is not None:
            extra_info = 'Cache'
            if not os.path.exists(cachefile):
                self.log('Fetching %s by url %s' % (sha1sum, url))
                if username:
                    opener = self.setupAuthHandler(
                        url, username, password).open
                else:
                    opener = urllib2.urlopen
                try:
                    f = opener(url)
                # Don't change this to URLError without thoroughly
                # testing for regressions. For now, just suppress
                # the PyLint warnings.
                # pylint: disable-msg=W0703
                except Exception, info:
                    extra_info = 'Error accessing Librarian: %s' % info
                    self.log(extra_info)
                else:
                    of = open(cachefile + '.tmp', "w")
                    # Upped for great justice to 256k
                    check_sum = hashlib.sha1()
                    for chunk in iter(lambda: f.read(256*1024), ''):
                        of.write(chunk)
                        check_sum.update(chunk)
                    of.close()
                    f.close()
                    extra_info = 'Download'
                    if check_sum.hexdigest() != sha1sum:
                        os.remove(cachefile + '.tmp')
                        extra_info = "Digests did not match, removing again!"
                    else:
                        os.rename(cachefile + '.tmp', cachefile)
                    self.log(extra_info)
        return (os.path.exists(cachefile), extra_info)

    def storeFile(self, path):
        """Store the content of the provided path in the file cache."""
        f = open(path)
        tmppath = self.cachePath("storeFile.tmp")
        of = open(tmppath, "w")
        try:
            sha1 = hashlib.sha1()
            for chunk in iter(lambda: f.read(256*1024), ''):
                sha1.update(chunk)
                of.write(chunk)
            sha1sum = sha1.hexdigest()
        finally:
            of.close()
            f.close()
        present, info = self.ensurePresent(sha1sum)
        if present:
            os.unlink(tmppath)
            return sha1sum
        os.rename(tmppath, self.cachePath(sha1sum))
        return sha1sum

    def addWaitingFile(self, path):
        """Add a file to the cache and store its details for reporting."""
        self.waitingfiles[os.path.basename(path)] = self.storeFile(path)

    def abort(self):
        """Abort the current build."""
        # XXX: dsilvers: 2005-01-21: Current abort mechanism doesn't wait
        # for abort to complete. This is potentially an issue in a heavy
        # load situation.
        if self.builderstatus == BuilderStatus.ABORTING:
            # This might happen if the master side restarts in the middle of
            # an abort cycle.
            self.log("Slave already ABORTING when asked to abort")
            return
        if self.builderstatus != BuilderStatus.BUILDING:
            # XXX: Should raise a known Fault so that the client can make
            # useful decisions about the error!
            raise ValueError("Slave is not BUILDING when asked to abort")
        self.manager.abort()
        self.builderstatus = BuilderStatus.ABORTING

    def clean(self):
        """Clean up pending files and reset the internal build state."""
        if self.builderstatus != BuilderStatus.WAITING:
            raise ValueError('Slave is not WAITING when asked to clean')
        for f in set(self.waitingfiles.values()):
            os.remove(self.cachePath(f))
        self.builderstatus = BuilderStatus.IDLE
        if self._log is not None:
            self._log.close()
            os.remove(self.cachePath("buildlog"))
            self._log = None
        self.waitingfiles = {}
        self.builddependencies = ""
        self.manager = None
        self.buildstatus = BuildStatus.OK

    def log(self, data):
        """Write the provided data to the log."""
        if self._log is not None:
            self._log.write(data)
            self._log.flush()
        if data.endswith("\n"):
            data = data[:-1]
        log.msg("Build log: " + data)

    def getLogTail(self):
        """Return the tail of the log.

        If the buildlog is not yet opened for writing (self._log is None),
        return a empty string.

        It safely tries to open the 'buildlog', if it doesn't exist, due to
        job cleanup or buildlog sanitization race-conditions, it also returns
        an empty string.

        When the 'buildlog' is present it return up to 2 KiB character of
        the end of the file.

        The returned content will be 'sanitized', see `_sanitizeURLs` for
        further information.
        """
        if self._log is None:
            return ""

        rlog = None
        try:
            try:
                rlog = open(self.cachePath("buildlog"), "r")
            except IOError:
                ret = ""
            else:
                # We rely on good OS practices that keep the file handler
                # usable once it's opened. So, if open() is ok, a subsequent
                # seek/tell/read will be safe.
                rlog.seek(0, os.SEEK_END)
                count = rlog.tell()
                if count > 2048:
                    count = 2048
                rlog.seek(-count, os.SEEK_END)
                ret = rlog.read(count)
        finally:
            if rlog is not None:
                rlog.close()

        if self.manager.is_archive_private:
            # This is a build in a private archive. We need to scrub
            # the URLs contained in the buildlog excerpt in order to
            # avoid leaking passwords.
            log_lines = ret.splitlines()

            # Please note: we are throwing away the first line (of the
            # excerpt to be scrubbed) because it may be cut off thus
            # thwarting the detection of embedded passwords.
            clean_content_iter = _sanitizeURLs(log_lines[1:])
            ret = '\n'.join(clean_content_iter)

        return ret

    def startBuild(self, manager):
        """Start a build with the provided BuildManager instance."""
        if self.builderstatus != BuilderStatus.IDLE:
            raise ValueError("Slave is not IDLE when asked to start building")
        self.manager = manager
        self.builderstatus = BuilderStatus.BUILDING
        self.emptyLog()

    def emptyLog(self):
        """Empty the log and start again."""
        if self._log is not None:
            self._log.close()
        self._log = open(self.cachePath("buildlog"), "w")

    def builderFail(self):
        """Cease building because the builder has a problem."""
        if self.builderstatus not in (BuilderStatus.BUILDING,
                                      BuilderStatus.ABORTING):
            raise ValueError("Slave is not BUILDING|ABORTING when set to "
                             "BUILDERFAIL")
        self.buildstatus = BuildStatus.BUILDERFAIL

    def chrootFail(self):
        """Cease building because the chroot could not be created or contained
        a set of package control files which couldn't upgrade themselves, or
        yet a lot of causes that imply the CHROOT is corrupted not the
        package.
        """
        if self.builderstatus != BuilderStatus.BUILDING:
            raise ValueError("Slave is not BUILDING when set to CHROOTFAIL")
        self.buildstatus = BuildStatus.CHROOTFAIL

    def buildFail(self):
        """Cease building because the package failed to build."""
        if self.builderstatus != BuilderStatus.BUILDING:
            raise ValueError("Slave is not BUILDING when set to PACKAGEFAIL")
        self.buildstatus = BuildStatus.PACKAGEFAIL

    def buildOK(self):
        """Having passed all possible failure states, mark a build as OK."""
        if self.builderstatus != BuilderStatus.BUILDING:
            raise ValueError("Slave is not BUILDING when set to OK")
        self.buildstatus = BuildStatus.OK

    def depFail(self, dependencies):
        """Cease building due to a dependency issue."""
        if self.builderstatus != BuilderStatus.BUILDING:
            raise ValueError("Slave is not BUILDING when set to DEPFAIL")
        self.buildstatus = BuildStatus.DEPFAIL
        self.builddependencies = dependencies

    def giveBack(self):
        """Give-back package due to a transient buildd/archive issue."""
        if self.builderstatus != BuilderStatus.BUILDING:
            raise ValueError("Slave is not BUILDING when set to GIVENBACK")
        self.buildstatus = BuildStatus.GIVENBACK

    def buildAborted(self):
        """Mark a build as aborted."""
        if self.builderstatus != BuilderStatus.ABORTING:
            raise ValueError("Slave is not ABORTING when set to ABORTED")
        if self.buildstatus != BuildStatus.BUILDERFAIL:
            self.buildstatus = BuildStatus.ABORTED

    def buildComplete(self):
        """Mark the build as complete and waiting interaction from the build
        daemon master.
        """
        if self.builderstatus == BuilderStatus.BUILDING:
            self.builderstatus = BuilderStatus.WAITING
        elif self.builderstatus == BuilderStatus.ABORTING:
            self.buildAborted()
            self.builderstatus = BuilderStatus.WAITING
        else:
            raise ValueError("Slave is not BUILDING|ABORTING when told build "
                             "is complete")

    def sanitizeBuildlog(self, log_path):
        """Removes passwords from buildlog URLs.

        Because none of the URLs to be processed are expected to span
        multiple lines and because build log files are potentially huge
        they will be processed line by line.

        :param log_path: The path to the buildlog file that is to be
            sanitized.
        :type log_path: ``str``
        """
        # First move the buildlog file that is to be sanitized out of
        # the way.
        unsanitized_path = self.cachePath(
            os.path.basename(log_path) + '.unsanitized')
        os.rename(log_path, unsanitized_path)

        # Open the unsanitized buildlog file for reading.
        unsanitized_file = open(unsanitized_path)

        # Open the file that will hold the resulting, sanitized buildlog
        # content for writing.
        sanitized_file = None

        try:
            sanitized_file = open(log_path, 'w')

            # Scrub the buildlog file line by line
            clean_content_iter = _sanitizeURLs(unsanitized_file)
            for line in clean_content_iter:
                sanitized_file.write(line)
        finally:
            # We're done with scrubbing, close the file handles.
            unsanitized_file.close()
            if sanitized_file is not None:
                sanitized_file.close()


class XMLRPCBuildDSlave(xmlrpc.XMLRPC):
    """XMLRPC build daemon slave management interface"""

    def __init__(self, config):
        xmlrpc.XMLRPC.__init__(self, allowNone=True)
        # The V1.0 new-style protocol introduces string-style protocol
        # versions of the form 'MAJOR.MINOR', the protocol is '1.0' for now
        # implying the presence of /filecache/ /filecache/buildlog and
        # the reduced and optimised XMLRPC interface.
        self.protocolversion = '1.0'
        self.slave = BuildDSlave(config)
        self._builders = {}
        cache = apt.Cache()
        try:
            try:
                self._version = cache["python-lpbuildd"].installed.version
            except AttributeError:
                # XXX cjwatson 2013-10-07: Remove when all builders are on
                # Ubuntu 10.04 LTS or newer.
                self._version = cache["python-lpbuildd"].installedVersion
        except KeyError:
            self._version = None
        log.msg("Initialized")

    def registerBuilder(self, builderclass, buildertag):
        self._builders[buildertag] = builderclass

    def xmlrpc_echo(self, *args):
        """Echo the argument back."""
        return args

    def xmlrpc_info(self):
        """Return the protocol version and the builder methods supported."""
        return (self.protocolversion, self.slave.getArch(),
                self._builders.keys())

    def xmlrpc_status(self):
        """Return the status of the build daemon, as a dictionary.

        Depending on the builder status we return differing amounts of data,
        but this always includes the builder status itself.
        """
        status = self.slave.builderstatus
        statusname = status.split('.')[-1]
        func = getattr(self, "status_" + statusname, None)
        if func is None:
            raise ValueError("Unknown status '%s'" % status)
        ret = {"builder_status": status}
        if self._version is not None:
            ret["builder_version"] = self._version
        ret.update(func())
        return ret

    def status_IDLE(self):
        """Handler for xmlrpc_status IDLE."""
        return {}

    def status_BUILDING(self):
        """Handler for xmlrpc_status BUILDING.

        Returns the build id and up to one kilobyte of log tail.
        """
        tail = self.slave.getLogTail()
        return {"build_id": self.buildid, "logtail": xmlrpclib.Binary(tail)}

    def status_WAITING(self):
        """Handler for xmlrpc_status WAITING.

        Returns the build id and the set of files waiting to be returned
        unless the builder failed in which case we return the buildstatus
        and the build id but no file set.
        """
        ret = {
            "build_status": self.slave.buildstatus,
            "build_id": self.buildid,
            }
        if self.slave.buildstatus in (BuildStatus.OK, BuildStatus.PACKAGEFAIL,
                                      BuildStatus.DEPFAIL):
            ret["filemap"] = self.slave.waitingfiles
            ret["dependencies"] = self.slave.builddependencies
        return ret

    def status_ABORTING(self):
        """Handler for xmlrpc_status ABORTING.

        This state means the builder is performing the ABORT command and is
        not able to do anything else than answer its status, so returns the
        build id only.
        """
        return {"build_id": self.buildid}

    def xmlrpc_ensurepresent(self, sha1sum, url, username, password):
        """Attempt to ensure the given file is present."""
        return self.slave.ensurePresent(sha1sum, url, username, password)

    def xmlrpc_abort(self):
        """Abort the current build."""
        self.slave.abort()
        return BuilderStatus.ABORTING

    def xmlrpc_clean(self):
        """Clean up the waiting files and reset the slave's internal state."""
        self.slave.clean()
        return BuilderStatus.IDLE

    def xmlrpc_build(self, buildid, builder, chrootsum, filemap, args):
        """Check if requested arguments are sane and initiate build procedure

        return a tuple containing: (<builder_status>, <info>)

        """
        # check requested builder
        if not builder in self._builders:
            extra_info = "%s not in %r" % (builder, self._builders.keys())
            return (BuilderStatus.UNKNOWNBUILDER, extra_info)
        # check requested chroot availability
        chroot_present, info = self.slave.ensurePresent(chrootsum)
        if not chroot_present:
            extra_info = """CHROOTSUM -> %s
            ***** INFO *****
            %s
            ****************
            """ % (chrootsum, info)
            return (BuilderStatus.UNKNOWNSUM, extra_info)
        # check requested files availability
        for filesum in filemap.itervalues():
            file_present, info = self.slave.ensurePresent(filesum)
            if not file_present:
                extra_info = """FILESUM -> %s
                ***** INFO *****
                %s
                ****************
                """ % (filesum, info)
                return (BuilderStatus.UNKNOWNSUM, extra_info)
        # check buildid sanity
        if buildid is None or buildid == "" or buildid == 0:
            raise ValueError(buildid)

        # builder is available, buildd is non empty,
        # filelist is consistent, chrootsum is available, let's initiate...
        self.buildid = buildid
        self.slave.startBuild(self._builders[builder](self.slave, buildid))
        self.slave.manager.initiate(filemap, chrootsum, args)
        return (BuilderStatus.BUILDING, buildid)



# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os

from lpbuildd.debian import DebianBuildManager, DebianBuildState


class TranslationTemplatesBuildState(DebianBuildState):
    INSTALL = "INSTALL"
    GENERATE = "GENERATE"


class TranslationTemplatesBuildManager(DebianBuildManager):
    """Generate translation templates from branch.

    This is the implementation of `TranslationTemplatesBuildJob`.  The
    latter runs on the master server; TranslationTemplatesBuildManager
    runs on the build slave.
    """

    initial_build_state = TranslationTemplatesBuildState.INSTALL

    def __init__(self, slave, buildid):
        super(TranslationTemplatesBuildManager, self).__init__(slave, buildid)
        self._generatepath = os.path.join(
            self._slavebin, "generate-translation-templates")
        self._resultname = slave._config.get(
            "translationtemplatesmanager", "resultarchive")

    def initiate(self, files, chroot, extra_args):
        """See `BuildManager`."""
        self._branch_url = extra_args['branch_url']
        self._chroot_path = os.path.join(
            self.home, 'build-' + self._buildid, 'chroot-autobuild')

        super(TranslationTemplatesBuildManager, self).initiate(
            files, chroot, extra_args)

    def doInstall(self):
        """Install packages required."""
        required_packages = [
            'bzr',
            'intltool',
            ]
        command = ['apt-get', 'install', '-y'] + required_packages
        chroot = ['sudo', 'chroot', self._chroot_path]
        self.runSubProcess('/usr/bin/sudo', chroot + command)

    # To satisfy DebianPackageManagers needs without having a misleading
    # method name here.
    doRunBuild = doInstall

    def doGenerate(self):
        """Generate templates."""
        command = [
            self._generatepath,
            self._buildid, self._branch_url, self._resultname]
        self.runSubProcess(self._generatepath, command)

    def gatherResults(self):
        """Gather the results of the build and add them to the file cache."""
        # The file is inside the chroot, in the home directory of the buildd
        # user. Should be safe to assume the home dirs are named identically.
        assert self.home.startswith('/'), "home directory must be absolute."

        path = os.path.join(
            self._chroot_path, self.home[1:], self._resultname)
        if os.access(path, os.F_OK):
            self._slave.addWaitingFile(path)

    def iterate_INSTALL(self, success):
        """Installation was done."""
        if success == 0:
            self._state = TranslationTemplatesBuildState.GENERATE
            self.doGenerate()
        else:
            if not self.alreadyfailed:
                self._slave.chrootFail()
                self.alreadyfailed = True
            self._state = TranslationTemplatesBuildState.UMOUNT
            self.doUnmounting()

    def iterate_GENERATE(self, success):
        """Template generation finished."""
        if success == 0:
            # It worked! Now let's bring in the harvest.
            self.gatherResults()
            self.doReapProcesses(self._state)
        else:
            if not self.alreadyfailed:
                self._slave.buildFail()
                self.alreadyfailed = True
            self.doReapProcesses(self._state)

    def iterateReap_GENERATE(self, success):
        """Finished reaping after template generation."""
        self._state = TranslationTemplatesBuildState.UMOUNT
        self.doUnmounting()



# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os
import shutil
import tempfile

from testtools import TestCase

from lpbuildd.livefs import (
    LiveFilesystemBuildManager,
    LiveFilesystemBuildState,
    )
from lpbuildd.tests.fakeslave import FakeSlave


class MockBuildManager(LiveFilesystemBuildManager):
    def __init__(self, *args, **kwargs):
        super(MockBuildManager, self).__init__(*args, **kwargs)
        self.commands = []
        self.iterators = []

    def runSubProcess(self, path, command, iterate=None):
        self.commands.append([path] + command)
        if iterate is None:
            iterate = self.iterate
        self.iterators.append(iterate)
        return 0


class TestLiveFilesystemBuildManagerIteration(TestCase):
    """Run LiveFilesystemBuildManager through its iteration steps."""
    def setUp(self):
        super(TestLiveFilesystemBuildManagerIteration, self).setUp()
        self.working_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.working_dir))
        slave_dir = os.path.join(self.working_dir, "slave")
        home_dir = os.path.join(self.working_dir, "home")
        for dir in (slave_dir, home_dir):
            os.mkdir(dir)
        self.slave = FakeSlave(slave_dir)
        self.buildid = "123"
        self.buildmanager = MockBuildManager(self.slave, self.buildid)
        self.buildmanager.home = home_dir
        self.buildmanager._cachepath = self.slave._cachepath
        self.build_dir = os.path.join(
            home_dir, "build-%s" % self.buildid, "chroot-autobuild", "build")

    def getState(self):
        """Retrieve build manager's state."""
        return self.buildmanager._state

    def startBuild(self):
        # The build manager's iterate() kicks off the consecutive states
        # after INIT.
        extra_args = {
            "project": "ubuntu",
            "series": "saucy",
            "pocket": "release",
            "arch_tag": "i386",
            }
        self.buildmanager.initiate({}, "chroot.tar.gz", extra_args)

        # Skip states that are done in DebianBuildManager to the state
        # directly before BUILD_LIVEFS.
        self.buildmanager._state = LiveFilesystemBuildState.UPDATE

        # BUILD_LIVEFS: Run the slave's payload to build the live filesystem.
        self.buildmanager.iterate(0)
        self.assertEqual(
            LiveFilesystemBuildState.BUILD_LIVEFS, self.getState())
        expected_command = [
            "sharepath/slavebin/buildlivefs", "buildlivefs", "--build-id",
            self.buildid, "--arch", "i386", "--project", "ubuntu",
            "--series", "saucy",
            ]
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled("chrootFail"))

    def test_iterate(self):
        # The build manager iterates a normal build from start to finish.
        self.startBuild()

        log_path = os.path.join(self.buildmanager._cachepath, "buildlog")
        log = open(log_path, "w")
        log.write("I am a build log.")
        log.close()

        os.makedirs(self.build_dir)
        manifest_path = os.path.join(self.build_dir, "livecd.ubuntu.manifest")
        manifest = open(manifest_path, "w")
        manifest.write("I am a manifest file.")
        manifest.close()

        # After building the package, reap processes.
        self.buildmanager.iterate(0)
        expected_command = [
            "sharepath/slavebin/scan-for-processes", "scan-for-processes",
            self.buildid,
            ]
        self.assertEqual(
            LiveFilesystemBuildState.BUILD_LIVEFS, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertNotEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled("buildFail"))
        self.assertEqual(
            [((manifest_path,), {})], self.slave.addWaitingFile.calls)

        # Control returns to the DebianBuildManager in the UMOUNT state.
        self.buildmanager.iterateReap(self.getState(), 0)
        expected_command = [
            "sharepath/slavebin/umount-chroot", "umount-chroot", self.buildid]
        self.assertEqual(LiveFilesystemBuildState.UMOUNT, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled("buildFail"))

    def test_omits_symlinks(self):
        # Symlinks in the build output are not included in gathered results.
        self.startBuild()

        log_path = os.path.join(self.buildmanager._cachepath, "buildlog")
        log = open(log_path, "w")
        log.write("I am a build log.")
        log.close()

        os.makedirs(self.build_dir)
        target_path = os.path.join(
            self.build_dir, "livecd.ubuntu.kernel-generic")
        target = open(target_path, "w")
        target.write("I am a kernel.")
        target.close()
        link_path = os.path.join(self.build_dir, "livecd.ubuntu.kernel")
        os.symlink("livecd.ubuntu.kernel-generic", link_path)

        self.buildmanager.iterate(0)
        self.assertEqual(
            [((target_path,), {})], self.slave.addWaitingFile.calls)



# Copyright 2010, 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os
import shutil
import tempfile

from testtools import TestCase

from lpbuildd.tests.fakeslave import FakeSlave
from lpbuildd.translationtemplates import (
    TranslationTemplatesBuildManager,
    TranslationTemplatesBuildState,
    )


class MockBuildManager(TranslationTemplatesBuildManager):
    def __init__(self, *args, **kwargs):
        super(MockBuildManager, self).__init__(*args, **kwargs)
        self.commands = []
        self.iterators = []

    def runSubProcess(self, path, command, iterate=None):
        self.commands.append([path]+command)
        if iterate is None:
            iterate = self.iterate
        self.iterators.append(iterate)
        return 0


class TestTranslationTemplatesBuildManagerIteration(TestCase):
    """Run TranslationTemplatesBuildManager through its iteration steps."""
    def setUp(self):
        super(TestTranslationTemplatesBuildManagerIteration, self).setUp()
        self.working_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.working_dir))
        slave_dir = os.path.join(self.working_dir, 'slave')
        home_dir = os.path.join(self.working_dir, 'home')
        for dir in (slave_dir, home_dir):
            os.mkdir(dir)
        self.slave = FakeSlave(slave_dir)
        self.buildid = '123'
        self.buildmanager = MockBuildManager(self.slave, self.buildid)
        self.buildmanager.home = home_dir
        self.chrootdir = os.path.join(
            home_dir, 'build-%s' % self.buildid, 'chroot-autobuild')

    def getState(self):
        """Retrieve build manager's state."""
        return self.buildmanager._state

    def test_iterate(self):
        # Two iteration steps are specific to this build manager.
        url = 'lp:~my/branch'
        # The build manager's iterate() kicks off the consecutive states
        # after INIT.
        self.buildmanager.initiate({}, 'chroot.tar.gz', {'branch_url': url})

        # Skip states that are done in DebianBuildManager to the state
        # directly before INSTALL.
        self.buildmanager._state = TranslationTemplatesBuildState.UPDATE

        # INSTALL: Install additional packages needed for this job into
        # the chroot.
        self.buildmanager.iterate(0)
        self.assertEqual(
            TranslationTemplatesBuildState.INSTALL, self.getState())
        expected_command = [
            '/usr/bin/sudo',
            'sudo', 'chroot', self.chrootdir,
            'apt-get',
            ]
        self.assertEqual(expected_command, self.buildmanager.commands[-1][:5])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])

        # GENERATE: Run the slave's payload, the script that generates
        # templates.
        self.buildmanager.iterate(0)
        self.assertEqual(
            TranslationTemplatesBuildState.GENERATE, self.getState())
        expected_command = [
            'sharepath/slavebin/generate-translation-templates',
            'sharepath/slavebin/generate-translation-templates',
            self.buildid, url, 'resultarchive'
            ]
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('chrootFail'))

        outfile_path = os.path.join(
            self.chrootdir, self.buildmanager.home[1:],
            self.buildmanager._resultname)
        os.makedirs(os.path.dirname(outfile_path))

        outfile = open(outfile_path, 'w')
        outfile.write("I am a template tarball. Seriously.")
        outfile.close()

        # After generating templates, reap processes.
        self.buildmanager.iterate(0)
        expected_command = [
            'sharepath/slavebin/scan-for-processes', 'scan-for-processes',
            self.buildid,
            ]
        self.assertEqual(
            TranslationTemplatesBuildState.GENERATE, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertNotEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('buildFail'))
        self.assertEqual(
            [((outfile_path,), {})], self.slave.addWaitingFile.calls)

        # The control returns to the DebianBuildManager in the UMOUNT state.
        self.buildmanager.iterateReap(self.getState(), 0)
        expected_command = [
            'sharepath/slavebin/umount-chroot', 'umount-chroot', self.buildid,
            ]
        self.assertEqual(
            TranslationTemplatesBuildState.UMOUNT, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('buildFail'))

    def test_iterate_fail_INSTALL(self):
        # See that a failing INSTALL is handled properly.
        url = 'lp:~my/branch'
        # The build manager's iterate() kicks off the consecutive states
        # after INIT.
        self.buildmanager.initiate({}, 'chroot.tar.gz', {'branch_url': url})

        # Skip states to the INSTALL state.
        self.buildmanager._state = TranslationTemplatesBuildState.INSTALL

        # The buildmanager fails and iterates to the UMOUNT state.
        self.buildmanager.iterate(-1)
        self.assertEqual(
            TranslationTemplatesBuildState.UMOUNT, self.getState())
        expected_command = [
            'sharepath/slavebin/umount-chroot', 'umount-chroot', self.buildid,
            ]
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertTrue(self.slave.wasCalled('chrootFail'))

    def test_iterate_fail_GENERATE(self):
        # See that a failing GENERATE is handled properly.
        url = 'lp:~my/branch'
        # The build manager's iterate() kicks off the consecutive states
        # after INIT.
        self.buildmanager.initiate({}, 'chroot.tar.gz', {'branch_url': url})

        # Skip states to the INSTALL state.
        self.buildmanager._state = TranslationTemplatesBuildState.GENERATE

        # The buildmanager fails and reaps processes.
        self.buildmanager.iterate(-1)
        expected_command = [
            'sharepath/slavebin/scan-for-processes', 'scan-for-processes',
            self.buildid,
            ]
        self.assertEqual(
            TranslationTemplatesBuildState.GENERATE, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertNotEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertTrue(self.slave.wasCalled('buildFail'))

        # The buildmanager iterates to the UMOUNT state.
        self.buildmanager.iterateReap(self.getState(), 0)
        self.assertEqual(
            TranslationTemplatesBuildState.UMOUNT, self.getState())
        expected_command = [
            'sharepath/slavebin/umount-chroot', 'umount-chroot', self.buildid
            ]
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])



# Copyright 2009, 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

# Authors: Daniel Silverstone <daniel.silverstone@canonical.com>
#      and Adam Conrad <adam.conrad@canonical.com>

# Buildd Slave sbuild manager implementation

__metaclass__ = type

import os
import re
import signal

from twisted.python import log

from lpbuildd.slave import (
    BuildManager,
    )


class DebianBuildState:
    """States for the DebianBuildManager."""
    INIT = "INIT"
    UNPACK = "UNPACK"
    MOUNT = "MOUNT"
    SOURCES = "SOURCES"
    UPDATE = "UPDATE"
    UMOUNT = "UMOUNT"
    CLEANUP = "CLEANUP"


class DebianBuildManager(BuildManager):
    """Base behaviour for Debian chrooted builds."""

    def __init__(self, slave, buildid, **kwargs):
        BuildManager.__init__(self, slave, buildid, **kwargs)
        self._updatepath = os.path.join(self._slavebin, "update-debian-chroot")
        self._sourcespath = os.path.join(
            self._slavebin, "override-sources-list")
        self._cachepath = slave._config.get("slave", "filecache")
        self._state = DebianBuildState.INIT
        slave.emptyLog()
        self.alreadyfailed = False

    @property
    def initial_build_state(self):
        raise NotImplementedError()

    def initiate(self, files, chroot, extra_args):
        """Initiate a build with a given set of files and chroot."""

        self.arch_tag = extra_args.get('arch_tag', self._slave.getArch())
        self.sources_list = extra_args.get('archives')

        BuildManager.initiate(self, files, chroot, extra_args)

    def doSourcesList(self):
        """Override apt/sources.list.

        Mainly used for PPA builds.
        """
        args = ["override-sources-list", self._buildid]
        args.extend(self.sources_list)
        self.runSubProcess(self._sourcespath, args)

    def doUpdateChroot(self):
        """Perform the chroot upgrade."""
        self.runSubProcess(
            self._updatepath,
            ["update-debian-chroot", self._buildid, self.arch_tag])

    def doRunBuild(self):
        """Run the main build process.

        Subclasses must override this.
        """
        raise NotImplementedError()

    @staticmethod
    def _parseChangesFile(linesIter):
        """A generator that iterates over files listed in a changes file.

        :param linesIter: an iterable of lines in a changes file.
        """
        seenfiles = False
        for line in linesIter:
            if line.endswith("\n"):
                line = line[:-1]
            if not seenfiles and line.startswith("Files:"):
                seenfiles = True
            elif seenfiles:
                if not line.startswith(' '):
                    break
                filename = line.split(' ')[-1]
                yield filename

    def getChangesFilename(self):
        changes = self._dscfile[:-4] + "_" + self.arch_tag + ".changes"
        return get_build_path(self.home, self._buildid, changes)

    def gatherResults(self):
        """Gather the results of the build and add them to the file cache.

        The primary file we care about is the .changes file. We key from there.
        """
        path = self.getChangesFilename()
        self._slave.addWaitingFile(path)

        chfile = open(path, "r")
        try:
            for fn in self._parseChangesFile(chfile):
                self._slave.addWaitingFile(
                    get_build_path(self.home, self._buildid, fn))
        finally:
            chfile.close()

    def iterate(self, success):
        # When a Twisted ProcessControl class is killed by SIGTERM,
        # which we call 'build process aborted', 'None' is returned as
        # exit_code.
        if self.alreadyfailed and success == 0:
            # We may have been aborted in between subprocesses; pretend that
            # we were terminated by a signal, which is close enough.
            success = 128 + signal.SIGKILL
        log.msg("Iterating with success flag %s against stage %s"
                % (success, self._state))
        func = getattr(self, "iterate_" + self._state, None)
        if func is None:
            raise ValueError, "Unknown internal state " + self._state
        func(success)

    def iterateReap(self, state, success):
        log.msg("Iterating with success flag %s against stage %s after "
                "reaping processes"
                % (success, state))
        func = getattr(self, "iterateReap_" + state, None)
        if func is None:
            raise ValueError, "Unknown internal post-reap state " + state
        func(success)

    def iterate_INIT(self, success):
        """Just finished initializing the build."""
        if success != 0:
            if not self.alreadyfailed:
                # The init failed, can't fathom why that would be...
                self._slave.builderFail()
                self.alreadyfailed = True
            self._state = DebianBuildState.CLEANUP
            self.doCleanup()
        else:
            self._state = DebianBuildState.UNPACK
            self.doUnpack()

    def iterate_UNPACK(self, success):
        """Just finished unpacking the tarball."""
        if success != 0:
            if not self.alreadyfailed:
                # The unpack failed for some reason...
                self._slave.chrootFail()
                self.alreadyfailed = True
            self._state = DebianBuildState.CLEANUP
            self.doCleanup()
        else:
            self._state = DebianBuildState.MOUNT
            self.doMounting()

    def iterate_MOUNT(self, success):
        """Just finished doing the mounts."""
        if success != 0:
            if not self.alreadyfailed:
                self._slave.chrootFail()
                self.alreadyfailed = True
            self._state = DebianBuildState.UMOUNT
            self.doUnmounting()
        else:
            if self.sources_list is not None:
                self._state = DebianBuildState.SOURCES
                self.doSourcesList()
            else:
                self._state = DebianBuildState.UPDATE
                self.doUpdateChroot()

    def searchLogContents(self, patterns_and_flags,
                          stop_patterns_and_flags=[]):
        """Search for any of a list of regex patterns in the build log.

        The build log is matched using a sliding window, which avoids having
        to read the whole file into memory at once but requires that matches
        be no longer than the chunk size (currently 256KiB).

        If any of the stop patterns are matched, the search stops
        immediately without reading the rest of the file.

        :return: A tuple of the regex pattern that matched and the match
            object, or (None, None).
        """
        chunk_size = 256 * 1024
        regexes = [
            re.compile(pattern, flags)
            for pattern, flags in patterns_and_flags]
        stop_regexes = [
            re.compile(pattern, flags)
            for pattern, flags in stop_patterns_and_flags]
        buildlog = open(os.path.join(self._cachepath, "buildlog"))
        try:
            window = ""
            chunk = buildlog.read(chunk_size)
            while chunk:
                window += chunk
                for regex in regexes:
                    match = regex.search(window)
                    if match is not None:
                        return regex.pattern, match
                for regex in stop_regexes:
                    if regex.search(window) is not None:
                        return None, None
                if len(window) > chunk_size:
                    window = window[chunk_size:]
                chunk = buildlog.read(chunk_size)
        finally:
            buildlog.close()
        return None, None

    def iterate_SOURCES(self, success):
        """Just finished overwriting sources.list."""
        if success != 0:
            if not self.alreadyfailed:
                self._slave.chrootFail()
                self.alreadyfailed = True
            self.doReapProcesses(self._state)
        else:
            self._state = DebianBuildState.UPDATE
            self.doUpdateChroot()

    def iterateReap_SOURCES(self, success):
        """Just finished reaping after failure to overwrite sources.list."""
        self._state = DebianBuildState.UMOUNT
        self.doUnmounting()

    def iterate_UPDATE(self, success):
        """Just finished updating the chroot."""
        if success != 0:
            if not self.alreadyfailed:
                self._slave.chrootFail()
                self.alreadyfailed = True
            self.doReapProcesses(self._state)
        else:
            self._state = self.initial_build_state
            self.doRunBuild()

    def iterateReap_UPDATE(self, success):
        """Just finished reaping after failure to update the chroot."""
        self._state = DebianBuildState.UMOUNT
        self.doUnmounting()

    def iterate_UMOUNT(self, success):
        """Just finished doing the unmounting."""
        if success != 0:
            if not self.alreadyfailed:
                self._slave.builderFail()
                self.alreadyfailed = True
        self._state = DebianBuildState.CLEANUP
        self.doCleanup()

    def iterate_CLEANUP(self, success):
        """Just finished the cleanup."""
        if success != 0:
            if not self.alreadyfailed:
                self._slave.builderFail()
                self.alreadyfailed = True
        else:
            # Successful clean
            if not self.alreadyfailed:
                self._slave.buildOK()
        self._slave.buildComplete()

    def abortReap(self):
        """Abort by killing all processes in the chroot, as hard as we can.

        Overridden here to handle state management.
        """
        self.doReapProcesses(self._state, notify=False)


def get_build_path(home, build_id, *extra):
    """Generate a path within the build directory.

    :param home: the user's home directory.
    :param build_id: the build id to use.
    :param extra: the extra path segments within the build directory.
    :return: the generated path.
    """
    return os.path.join(home, "build-" + build_id, *extra)



# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os
import shutil

from lpbuildd.debian import (
    DebianBuildManager,
    DebianBuildState,
    get_build_path,
    )


RETCODE_SUCCESS = 0
RETCODE_FAILURE_INSTALL = 200
RETCODE_FAILURE_BUILD = 201


class LiveFilesystemBuildState(DebianBuildState):
    BUILD_LIVEFS = "BUILD_LIVEFS"


class LiveFilesystemBuildManager(DebianBuildManager):
    """Build a live filesystem."""

    initial_build_state = LiveFilesystemBuildState.BUILD_LIVEFS

    def __init__(self, slave, buildid, **kwargs):
        DebianBuildManager.__init__(self, slave, buildid, **kwargs)
        self.build_livefs_path = os.path.join(self._slavebin, "buildlivefs")

    def initiate(self, files, chroot, extra_args):
        """Initiate a build with a given set of files and chroot."""
        self.build_path = get_build_path(
            self.home, self._buildid, "chroot-autobuild", "build")
        if os.path.isdir(self.build_path):
            shutil.rmtree(self.build_path)

        self.subarch = extra_args.get("subarch")
        self.project = extra_args["project"]
        self.subproject = extra_args.get("subproject")
        self.series = extra_args["series"]
        self.pocket = extra_args["pocket"]
        self.datestamp = extra_args.get("datestamp")
        self.image_format = extra_args.get("image_format")
        self.locale = extra_args.get("locale")
        self.extra_ppas = extra_args.get("extra_ppas", [])

        super(LiveFilesystemBuildManager, self).initiate(
            files, chroot, extra_args)

    def doRunBuild(self):
        """Run the process to build the live filesystem."""
        args = [
            "buildlivefs",
            "--build-id", self._buildid,
            "--arch", self.arch_tag,
            ]
        if self.subarch:
            args.extend(["--subarch", self.subarch])
        args.extend(["--project", self.project])
        if self.subproject:
            args.extend(["--subproject", self.subproject])
        args.extend(["--series", self.series])
        if self.datestamp:
            args.extend(["--datestamp", self.datestamp])
        if self.image_format:
            args.extend(["--image-format", self.image_format])
        if self.pocket == "proposed":
            args.append("--proposed")
        if self.locale:
            args.extend(["--locale", self.locale])
        for ppa in self.extra_ppas:
            args.extend(["--extra-ppa", ppa])
        self.runSubProcess(self.build_livefs_path, args)

    def iterate_BUILD_LIVEFS(self, retcode):
        """Finished building the live filesystem."""
        if retcode == RETCODE_SUCCESS:
            self.gatherResults()
            print("Returning build status: OK")
        elif (retcode >= RETCODE_FAILURE_INSTALL and
              retcode <= RETCODE_FAILURE_BUILD):
            if not self.alreadyfailed:
                self._slave.buildFail()
                print("Returning build status: Build failed.")
            self.alreadyfailed = True
        else:
            if not self.alreadyfailed:
                self._slave.builderFail()
                print("Returning build status: Builder failed.")
            self.alreadyfailed = True
        self.doReapProcesses(self._state)

    def iterateReap_BUILD_LIVEFS(self, retcode):
        """Finished reaping after building the live filesystem."""
        self._state = DebianBuildState.UMOUNT
        self.doUnmounting()

    def gatherResults(self):
        """Gather the results of the build and add them to the file cache."""
        for entry in sorted(os.listdir(self.build_path)):
            path = os.path.join(self.build_path, entry)
            if entry.startswith("livecd.") and not os.path.islink(path):
                self._slave.addWaitingFile(path)



# Copyright 2009-2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Functions to build PO templates on the build slave."""

__metaclass__ = type
__all__ = [
    'check_potfiles_in',
    'generate_pot',
    'generate_pots',
    'get_translation_domain',
    'find_intltool_dirs',
    'find_potfiles_in',
    ]

from contextlib import contextmanager
import errno
import os.path
import re
from subprocess import call


def find_potfiles_in():
    """Search the current directory and its subdirectories for POTFILES.in.

    :returns: A list of names of directories that contain a file POTFILES.in.
    """
    result_dirs = []
    for dirpath, dirnames, dirfiles in os.walk("."):
        if "POTFILES.in" in dirfiles:
            result_dirs.append(dirpath)
    return result_dirs


def check_potfiles_in(path):
    """Check if the files listed in the POTFILES.in file exist.

    Running 'intltool-update -m' will perform this check and also take a
    possible POTFILES.skip into account. It stores details about 'missing'
    (files that should be in POTFILES.in) and 'notexist'ing files (files
    that are listed in POTFILES.in but don't exist) in files which are
    named accordingly. These files are removed before the run.

    We don't care about files missing from POTFILES.in but want to know if
    all listed files exist. The presence of the 'notexist' file tells us
    that.

    :param path: The directory where POTFILES.in resides.
    :returns: False if the directory does not exist, if an error occurred
        when executing intltool-update or if files are missing from
        POTFILES.in. True if all went fine and all files in POTFILES.in
        actually exist.  
    """
    current_path = os.getcwd()

    try:
        os.chdir(path)
    except OSError, e:
        # Abort nicely if the directory does not exist.
        if e.errno == errno.ENOENT:
            return False
        raise
    try:
        # Remove stale files from a previous run of intltool-update -m.
        for unlink_name in ['missing', 'notexist']:
            try:
                os.unlink(unlink_name)
            except OSError, e:
                # It's ok if the files are missing.
                if e.errno != errno.ENOENT:
                    raise
        devnull = open("/dev/null", "w")
        returncode = call(
            ["/usr/bin/intltool-update", "-m"],
            stdout=devnull, stderr=devnull)
        devnull.close()
    finally:
        os.chdir(current_path)

    if returncode != 0:
        # An error occurred when executing intltool-update.
        return False

    notexist = os.path.join(path, "notexist")
    return not os.access(notexist, os.R_OK)


def find_intltool_dirs():
    """Search for directories with intltool structure.

    The current directory and its subdiretories are searched. An 'intltool
    structure' is a directory that contains a POFILES.in file and where all
    files listed in that POTFILES.in do actually exist. The latter
    condition makes sure that the file is not stale.

    :returns: A list of directory names.
    """
    return sorted(filter(check_potfiles_in, find_potfiles_in()))


def _get_AC_PACKAGE_NAME(config_file):
    """Get the value of AC_PACKAGE_NAME from function parameters.

    The value of AC_PACKAGE_NAME is either the first or the fourth
    parameter of the AC_INIT call if it is called with at least two
    parameters.
    """
    params = config_file.getFunctionParams("AC_INIT")
    if params is None or len(params) < 2:
        return None
    if len(params) < 4:
        return params[0]
    else:
        return params[3]


def _try_substitution(config_files, varname, substitution):
    """Try to find a substitution in the config files.

    :returns: The completed substitution or None if none was found.
    """
    subst_value = None
    if varname == substitution.name:
        # Do not look for the same name in the current file.
        config_files = config_files[:-1]
    for config_file in reversed(config_files):
        subst_value = config_file.getVariable(substitution.name)
        if subst_value is not None:
            # Substitution found.
            break
    else:
        # No substitution found.
        return None
    return substitution.replace(subst_value)


def get_translation_domain(dirname):
    """Get the translation domain for this PO directory.

    Imitates some of the behavior of intltool-update to find out which
    translation domain the build environment provides. The domain is usually
    defined in the GETTEXT_PACKAGE variable in one of the build files. Another
    variant is DOMAIN in the Makevars file. This function goes through the
    ordered list of these possible locations, top to bottom, and tries to
    find a valid value. Since the same variable name may be defined in
    multiple files (usually configure.ac and Makefile.in.in), it needs to
    keep trying with the next file, until it finds the most specific
    definition.

    If the found value contains a substitution, either autoconf style (@...@)
    or make style ($(...)), the search is continued in the same file and back
    up the list of files, now searching for the substitution. Multiple
    substitutions or multi-level substitutions are not supported.
    """
    locations = [
        ('../configure.ac', 'GETTEXT_PACKAGE', True),
        ('../configure.in', 'GETTEXT_PACKAGE', True),
        ('Makefile.in.in', 'GETTEXT_PACKAGE', False),
        ('Makevars', 'DOMAIN', False),
    ]
    value = None
    substitution = None
    config_files = []
    for filename, varname, keep_trying in locations:
        path = os.path.join(dirname, filename)
        if not os.access(path, os.R_OK):
            # Skip non-existent files.
            continue
        config_files.append(ConfigFile(path))
        new_value = config_files[-1].getVariable(varname)
        if new_value is not None:
            value = new_value
            if value == "AC_PACKAGE_NAME":
                value = _get_AC_PACKAGE_NAME(config_files[-1])
            else:
                # Check if the value needs a substitution.
                substitution = Substitution.get(value)
                if substitution is not None:
                    # Try to substitute with value.
                    value = _try_substitution(
                        config_files, varname, substitution)
                    if value is None:
                        # No substitution found; the setup is broken.
                        break
        if value is not None and not keep_trying:
            # A value has been found.
            break
    return value


@contextmanager
def chdir(directory):
    cwd = os.getcwd()
    os.chdir(directory)
    yield
    os.chdir(cwd)


def generate_pot(podir, domain):
    """Generate one PO template using intltool.

    Although 'intltool-update -p' can try to find out the translation domain
    we trust our own code more on this one and simply specify the domain.
    Also, the man page for 'intltool-update' states that the '-g' option
    "has an additional effect: the name of current working directory is no
    more  limited  to 'po' or 'po-*'." We don't want that limit either.

    :param podir: The PO directory in which to build template.
    :param domain: The translation domain to use as the name of the template.
      If it is None or empty, 'messages.pot' will be used.
    :return: True if generation succeeded.
    """
    if domain is None or domain.strip() == "":
        domain = "messages"
    with chdir(podir):
        with open("/dev/null", "w") as devnull:
            returncode = call(
                ["/usr/bin/intltool-update", "-p", "-g", domain],
                stdout=devnull, stderr=devnull)
    return returncode == 0


def generate_pots(package_dir='.'):
    """Top-level function to generate all PO templates in a package."""
    potpaths = []
    with chdir(package_dir):
        for podir in find_intltool_dirs():
            domain = get_translation_domain(podir)
            if generate_pot(podir, domain):
                potpaths.append(os.path.join(podir, domain + ".pot"))
    return potpaths


class ConfigFile(object):
    """Represent a config file and return variables defined in it."""

    def __init__(self, file_or_name):
        if isinstance(file_or_name, basestring):
            conf_file = file(file_or_name)
        else:
            conf_file = file_or_name
        self.content = conf_file.read()

    def _stripQuotes(self, identifier):
        """Strip surrounding quotes from `identifier`, if present.

        :param identifier: a string, possibly surrounded by matching
            'single,' "double," or [bracket] quotes.
        :return: `identifier` but with the outer pair of matching quotes
            removed, if they were there.
        """
        if len(identifier) < 2:
            return identifier

        quote_pairs = [
            ('"', '"'),
            ("'", "'"),
            ("[", "]"),
            ]
        for (left, right) in quote_pairs:
            if identifier.startswith(left) and identifier.endswith(right):
                return identifier[1:-1]

        return identifier

    def getVariable(self, name):
        """Search the file for a variable definition with this name."""
        pattern = re.compile(
            "^%s[ \t]*=[ \t]*([^\s]*)" % re.escape(name), re.M)
        result = pattern.search(self.content)
        if result is None:
            return None
        return self._stripQuotes(result.group(1))

    def getFunctionParams(self, name):
        """Search file for a function call with this name, return parameters.
        """
        pattern = re.compile("^%s\(([^)]*)\)" % re.escape(name), re.M)
        result = pattern.search(self.content)
        if result is None:
            return None
        else:
            return [
                self._stripQuotes(param.strip())
                for param in result.group(1).split(',')
                ]


class Substitution(object):
    """Find and replace substitutions.

    Variable texts may contain other variables which should be substituted
    for their value. These are either marked by surrounding @ signs (autoconf
    style) or preceded by a $ sign with optional () (make style).

    This class identifies a single such substitution in a variable text and
    extract the name of the variable who's value is to be inserted. It also
    facilitates the actual replacement so that caller does not have to worry
    about the substitution style that is being used.
    """

    autoconf_pattern = re.compile("@([^@]+)@")
    makefile_pattern = re.compile("\$\(?([^\s\)]+)\)?")

    @staticmethod
    def get(variabletext):
        """Factory method.

        Creates a Substitution instance and checks if it found a substitution.

        :param variabletext: A variable value with possible substitution.
        :returns: A Substitution object or None if no substitution was found.
        """
        subst = Substitution(variabletext)
        if subst.name is not None:
            return subst
        return None

    def _searchForPatterns(self):
        """Search for all the available patterns in variable text."""
        result = self.autoconf_pattern.search(self.text)
        if result is None:
            result = self.makefile_pattern.search(self.text)
        return result

    def __init__(self, variabletext):
        """Extract substitution name from variable text."""
        self.text = variabletext
        self.replaced = False
        result = self._searchForPatterns()
        if result is None:
            self._replacement = None
            self.name = None
        else:
            self._replacement = result.group(0)
            self.name = result.group(1)

    def replace(self, value):
        """Return a copy of the variable text with the substitution resolved.
        """
        self.replaced = True
        return self.text.replace(self._replacement, value)



# Copyright 2009-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type
__all__ = [
    'BuilddTestCase',
    ]

import os
import shutil
import tempfile
import unittest
from ConfigParser import SafeConfigParser

from fixtures import EnvironmentVariable
from txfixtures.tachandler import TacTestFixture

from lpbuildd.slave import BuildDSlave


test_conffile = os.path.join(
    os.path.dirname(__file__), 'buildd-slave-test.conf')


class MockBuildManager(object):
    """Mock BuildManager class.

    Only implements 'is_archive_private' as False.
    """
    is_archive_private = False


class BuilddTestCase(unittest.TestCase):
    """Unit tests for logtail mechanisms."""

    def setUp(self):
        """Setup a BuildDSlave using the test config."""
        conf = SafeConfigParser()
        conf.read(test_conffile)
        conf.set("slave", "filecache", tempfile.mkdtemp())

        self.slave = BuildDSlave(conf)
        self.slave._log = True
        self.slave.manager = MockBuildManager()

        self.here = os.path.abspath(os.path.dirname(__file__))

    def tearDown(self):
        """Remove the 'filecache' directory used for the tests."""
        shutil.rmtree(self.slave._cachepath)

    def makeLog(self, size):
        """Inject data into the default buildlog file."""
        f = open(self.slave.cachePath('buildlog'), 'w')
        f.write("x" * size)
        f.close()


class BuilddSlaveTestSetup(TacTestFixture):
    r"""Setup BuildSlave for use by functional tests

    >>> fixture = BuilddSlaveTestSetup()
    >>> fixture.setUp()

    Make sure the server is running

    >>> import xmlrpclib
    >>> s = xmlrpclib.Server('http://localhost:8221/rpc/')
    >>> s.echo('Hello World')
    ['Hello World']
    >>> fixture.tearDown()

    Again for luck !

    >>> fixture.setUp()
    >>> s = xmlrpclib.Server('http://localhost:8221/rpc/')

    >>> s.echo('Hello World')
    ['Hello World']

    >>> info = s.info()
    >>> len(info)
    3
    >>> print info[:2]
    ['1.0', 'i386']

    >>> for buildtype in sorted(info[2]):
    ...     print buildtype
    binarypackage
    debian
    sourcepackagerecipe
    translation-templates

    >>> s.status()["builder_status"]
    'BuilderStatus.IDLE'

    >>> fixture.tearDown()
    """
    def setUpRoot(self):
        """Recreate empty root directory to avoid problems."""
        if os.path.exists(self.root):
            shutil.rmtree(self.root)
        os.mkdir(self.root)
        filecache = os.path.join(self.root, 'filecache')
        os.mkdir(filecache)
        self.useFixture(EnvironmentVariable('HOME', self.root))
        self.useFixture(
            EnvironmentVariable('BUILDD_SLAVE_CONFIG', test_conffile))
        # XXX cprov 2005-05-30:
        # When we are about running it seriously we need :
        # * install sbuild package
        # * to copy the scripts for sbuild
        self.addCleanup(shutil.rmtree, self.root)

    @property
    def root(self):
        return '/var/tmp/buildd'

    @property
    def tacfile(self):
        return os.path.abspath(os.path.join(
            os.path.dirname(__file__),
            os.path.pardir,
            os.path.pardir,
            'buildd-slave.tac'
            ))

    @property
    def pidfile(self):
        return os.path.join(self.root, 'build-slave.pid')

    @property
    def logfile(self):
        return '/var/tmp/build-slave.log'

    @property
    def daemon_port(self):
        # This must match buildd-slave-test.conf.
        return 8221



# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type
__all__ = [
    'FakeMethod',
    'FakeSlave',
    ]

import os


class FakeMethod:
    """Catch any function or method call, and record the fact.

    Use this for easy stubbing.  The call operator can return a fixed
    value, or raise a fixed exception object.

    This is useful when unit-testing code that does things you don't
    want to integration-test, e.g. because it wants to talk to remote
    systems.
    """

    def __init__(self, result=None, failure=None):
        """Set up a fake function or method.

        :param result: Value to return.
        :param failure: Exception to raise.
        """
        self.result = result
        self.failure = failure

        # A log of arguments for each call to this method.
        self.calls = []

    def __call__(self, *args, **kwargs):
        """Catch an invocation to the method.

        Increment `call_count`, and adds the arguments to `calls`.

        Accepts any and all parameters.  Raises the failure passed to
        the constructor, if any; otherwise, returns the result value
        passed to the constructor.
        """
        self.calls.append((args, kwargs))

        if self.failure is None:
            return self.result
        else:
            # pylint thinks this raises None, which is clearly not
            # possible.  That's why this test disables pylint message
            # E0702.
            raise self.failure

    @property
    def call_count(self):
        return len(self.calls)

    def extract_args(self):
        """Return just the calls' positional-arguments tuples."""
        return [args for args, kwargs in self.calls]

    def extract_kwargs(self):
        """Return just the calls' keyword-arguments dicts."""
        return [kwargs for args, kwargs in self.calls]


class FakeConfig:
    def get(self, section, key):
        return key


class FakeSlave:
    def __init__(self, tempdir):
        self._cachepath = tempdir
        self._config = FakeConfig()
        self.waitingfiles = {}
        for fake_method in (
            "storeFile", "addWaitingFile", "emptyLog", "log",
            "chrootFail", "buildFail", "builderFail", "depFail",
            ):
            setattr(self, fake_method, FakeMethod())

    def cachePath(self, file):
        return os.path.join(self._cachepath, file)

    def anyMethod(self, *args, **kwargs):
        pass

    def wasCalled(self, name):
        return getattr(self, name).call_count > 0

    def getArch(self):
        return 'i386'



# Copyright 2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os
import shutil

from lpbuildd.debian import (
    DebianBuildManager,
    DebianBuildState,
    get_build_path,
    )


RETCODE_SUCCESS = 0
RETCODE_FAILURE_INSTALL = 200
RETCODE_FAILURE_BUILD = 201


class SnapBuildState(DebianBuildState):
    BUILD_SNAP = "BUILD_SNAP"


class SnapBuildManager(DebianBuildManager):
    """Build a snap."""

    initial_build_state = SnapBuildState.BUILD_SNAP

    def __init__(self, slave, buildid, **kwargs):
        super(SnapBuildManager, self).__init__(slave, buildid, **kwargs)
        self.build_snap_path = os.path.join(self._slavebin, "buildsnap")

    def initiate(self, files, chroot, extra_args):
        """Initiate a build with a given set of files and chroot."""
        self.build_path = get_build_path(
            self.home, self._buildid, "chroot-autobuild", "build")
        if os.path.isdir(self.build_path):
            shutil.rmtree(self.build_path)

        self.name = extra_args["name"]
        self.branch = extra_args.get("branch")
        self.git_repository = extra_args.get("git_repository")
        self.git_path = extra_args.get("git_path")

        super(SnapBuildManager, self).initiate(files, chroot, extra_args)

    def doRunBuild(self):
        """Run the process to build the snap."""
        args = [
            "buildsnap",
            "--build-id", self._buildid,
            "--arch", self.arch_tag,
            ]
        if self.branch is not None:
            args.extend(["--branch", self.branch])
        if self.git_repository is not None:
            args.extend(["--git-repository", self.git_repository])
        if self.git_path is not None:
            args.extend(["--git-path", self.git_path])
        args.append(self.name)
        self.runSubProcess(self.build_snap_path, args)

    def iterate_BUILD_SNAP(self, retcode):
        """Finished building the snap."""
        if retcode == RETCODE_SUCCESS:
            self.gatherResults()
            print("Returning build status: OK")
        elif (retcode >= RETCODE_FAILURE_INSTALL and
              retcode <= RETCODE_FAILURE_BUILD):
            if not self.alreadyfailed:
                self._slave.buildFail()
                print("Returning build status: Build failed.")
            self.alreadyfailed = True
        else:
            if not self.alreadyfailed:
                self._slave.builderFail()
                print("Returning build status: Builder failed.")
            self.alreadyfailed = True
        self.doReapProcesses(self._state)

    def iterateReap_BUILD_SNAP(self, retcode):
        """Finished reaping after building the snap."""
        self._state = DebianBuildState.UMOUNT
        self.doUnmounting()

    def gatherResults(self):
        """Gather the results of the build and add them to the file cache."""
        output_path = os.path.join(self.build_path, self.name)
        if not os.path.exists(output_path):
            return
        for entry in sorted(os.listdir(output_path)):
            path = os.path.join(output_path, entry)
            if entry.endswith(".snap") and not os.path.islink(path):
                self._slave.addWaitingFile(path)



# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

from contextlib import contextmanager
import imp
import os
import shutil
import sys
import tempfile
from textwrap import dedent

from testtools import TestCase


@contextmanager
def disable_bytecode():
    original = sys.dont_write_bytecode
    sys.dont_write_bytecode = True
    yield
    sys.dont_write_bytecode = original


# By-hand import to avoid having to put .py suffixes on slave binaries.
with disable_bytecode():
    RecipeBuilder = imp.load_source("buildrecipe", "buildrecipe").RecipeBuilder


class TestRecipeBuilder(TestCase):
    def setUp(self):
        super(TestRecipeBuilder, self).setUp()
        self.save_env = dict(os.environ)
        self.home_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.home_dir))
        os.environ["HOME"] = self.home_dir
        self.build_id = "1"
        self.builder = RecipeBuilder(
            self.build_id, "Recipe Builder", "builder@example.org>", "grumpy",
            "grumpy", "main", "PPA")
        os.makedirs(self.builder.work_dir)

    def resetEnvironment(self):
        for key in set(os.environ.keys()) - set(self.save_env.keys()):
            del os.environ[key]
        for key, value in os.environ.items():
            if value != self.save_env[key]:
                os.environ[key] = self.save_env[key]
        for key in set(self.save_env.keys()) - set(os.environ.keys()):
            os.environ[key] = self.save_env[key]

    def tearDown(self):
        self.resetEnvironment()
        super(TestRecipeBuilder, self).tearDown()

    def test_makeDummyDsc(self):
        self.builder.source_dir_relative = os.path.join(
            self.builder.work_dir_relative, "tree", "foo")
        control_path = os.path.join(
            self.builder.work_dir, "tree", "foo", "debian", "control")
        os.makedirs(os.path.dirname(control_path))
        os.makedirs(self.builder.apt_dir)
        with open(control_path, "w") as control:
            print >>control, dedent("""\
                Source: foo
                Build-Depends: debhelper (>= 9~), libfoo-dev

                Package: foo
                Depends: ${shlibs:Depends}""")
        self.builder.makeDummyDsc("foo")
        with open(os.path.join(self.builder.apt_dir, "foo.dsc")) as dsc:
            self.assertEqual(
                dedent("""\
                    Format: 1.0
                    Source: foo
                    Architecture: any
                    Version: 99:0
                    Maintainer: invalid@example.org
                    Build-Depends: debhelper (>= 9~), libfoo-dev

                    """),
                dsc.read())

    def test_makeDummyDsc_comments(self):
        # apt_pkg.TagFile doesn't support comments, but python-debian's own
        # parser does.  Make sure we're using the right one.
        self.builder.source_dir_relative = os.path.join(
            self.builder.work_dir_relative, "tree", "foo")
        control_path = os.path.join(
            self.builder.work_dir, "tree", "foo", "debian", "control")
        os.makedirs(os.path.dirname(control_path))
        os.makedirs(self.builder.apt_dir)
        with open(control_path, "w") as control:
            print >>control, dedent("""\
                Source: foo
                Build-Depends: debhelper (>= 9~),
                               libfoo-dev,
                # comment line
                               pkg-config

                Package: foo
                Depends: ${shlibs:Depends}""")
        self.builder.makeDummyDsc("foo")
        with open(os.path.join(self.builder.apt_dir, "foo.dsc")) as dsc:
            self.assertEqual(
                dedent("""\
                    Format: 1.0
                    Source: foo
                    Architecture: any
                    Version: 99:0
                    Maintainer: invalid@example.org
                    Build-Depends: debhelper (>= 9~),
                                   libfoo-dev,
                                   pkg-config

                    """),
                dsc.read())

    def test_runAptFtparchive(self):
        os.makedirs(self.builder.apt_dir)
        with open(os.path.join(self.builder.apt_dir, "foo.dsc"), "w") as dsc:
            print >>dsc, dedent("""\
                Format: 1.0
                Source: foo
                Architecture: any
                Version: 99:0
                Maintainer: invalid@example.org
                Build-Depends: debhelper (>= 9~), libfoo-dev""")
        self.assertEqual(0, self.builder.runAptFtparchive())
        self.assertEqual(
            ["Release", "Sources", "Sources.bz2", "foo.dsc",
             "ftparchive.conf"],
            sorted(os.listdir(self.builder.apt_dir)))
        with open(os.path.join(self.builder.apt_dir, "Sources")) as sources:
            sources_text = sources.read()
            self.assertIn("Package: foo\n", sources_text)
            self.assertIn(
                "Build-Depends: debhelper (>= 9~), libfoo-dev\n", sources_text)

    # XXX cjwatson 2015-06-15: We should unit-test enableAptArchive and
    # installBuildDeps too, but it involves a lot of mocks.  For now,
    # integration testing is probably more useful.



# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os
import shutil
import tempfile
from textwrap import dedent

from debian.deb822 import PkgRelation
from testtools import TestCase
from testtools.matchers import (
    ContainsDict,
    Equals,
    Is,
    MatchesListwise,
    )
from twisted.internet.task import Clock

from lpbuildd.binarypackage import (
    BinaryPackageBuildManager,
    BinaryPackageBuildState,
    SBuildExitCodes,
    )
from lpbuildd.tests.fakeslave import (
    FakeMethod,
    FakeSlave,
    )


class MockTransport:
    def __init__(self):
        self.loseConnection = FakeMethod()
        self.signalProcess = FakeMethod()


class MockSubprocess:
    def __init__(self, path):
        self.path = path
        self.transport = MockTransport()


class MockBuildManager(BinaryPackageBuildManager):
    def __init__(self, *args, **kwargs):
        super(MockBuildManager, self).__init__(*args, **kwargs)
        self.commands = []
        self.iterators = []
        self.arch_indep = False

    def runSubProcess(self, path, command, iterate=None):
        self.commands.append([path]+command)
        if iterate is None:
            iterate = self.iterate
        self.iterators.append(iterate)
        self._subprocess = MockSubprocess(path)
        return 0


def write_file(path, content):
    if not os.path.isdir(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    with open(path, 'w') as f:
        f.write(content)


class TestBinaryPackageBuildManagerIteration(TestCase):
    """Run BinaryPackageBuildManager through its iteration steps."""
    def setUp(self):
        super(TestBinaryPackageBuildManagerIteration, self).setUp()
        self.working_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.working_dir))
        slave_dir = os.path.join(self.working_dir, 'slave')
        home_dir = os.path.join(self.working_dir, 'home')
        for dir in (slave_dir, home_dir):
            os.mkdir(dir)
        self.slave = FakeSlave(slave_dir)
        self.buildid = '123'
        self.clock = Clock()
        self.buildmanager = MockBuildManager(
            self.slave, self.buildid, reactor=self.clock)
        self.buildmanager.home = home_dir
        self.buildmanager._cachepath = self.slave._cachepath
        self.chrootdir = os.path.join(
            home_dir, 'build-%s' % self.buildid, 'chroot-autobuild')

    def getState(self):
        """Retrieve build manager's state."""
        return self.buildmanager._state

    def startBuild(self, dscname=''):
        # The build manager's iterate() kicks off the consecutive states
        # after INIT.
        self.buildmanager.initiate(
            {'foo_1.dsc': dscname}, 'chroot.tar.gz',
            {'distribution': 'ubuntu', 'suite': 'warty',
             'ogrecomponent': 'main'})

        os.makedirs(self.chrootdir)

        # Skip DebianBuildManager states to the state directly before
        # SBUILD.
        self.buildmanager._state = BinaryPackageBuildState.UPDATE

        # SBUILD: Build the package.
        self.buildmanager.iterate(0)
        self.assertState(
            BinaryPackageBuildState.SBUILD,
            [
            'sharepath/slavebin/sbuild-package', 'sbuild-package',
            self.buildid, 'i386', 'warty', '-c', 'chroot:autobuild',
            '--arch=i386', '--dist=warty', '--purge=never', '--nolog',
            'foo_1.dsc',
            ], True)
        self.assertFalse(self.slave.wasCalled('chrootFail'))

    def assertState(self, state, command, final):
        self.assertEqual(state, self.getState())
        self.assertEqual(command, self.buildmanager.commands[-1])
        if final:
            self.assertEqual(
                self.buildmanager.iterate, self.buildmanager.iterators[-1])
        else:
            self.assertNotEqual(
                self.buildmanager.iterate, self.buildmanager.iterators[-1])

    def assertScansSanely(self, exit_code):
        # After building the package, reap processes.
        self.buildmanager.iterate(exit_code)
        self.assertState(
            BinaryPackageBuildState.SBUILD,
            ['sharepath/slavebin/scan-for-processes', 'scan-for-processes',
             self.buildid], False)

    def assertUnmountsSanely(self):
        self.buildmanager.iterateReap(self.getState(), 0)
        self.assertState(
            BinaryPackageBuildState.UMOUNT,
            ['sharepath/slavebin/umount-chroot', 'umount-chroot',
             self.buildid], True)

    def test_iterate(self):
        # The build manager iterates a normal build from start to finish.
        self.startBuild()

        write_file(
            os.path.join(self.buildmanager._cachepath, 'buildlog'),
            "I am a build log.")
        changes_path = os.path.join(
            self.buildmanager.home, 'build-%s' % self.buildid,
            'foo_1_i386.changes')
        write_file(changes_path, "I am a changes file.")

        # After building the package, reap processes.
        self.assertScansSanely(SBuildExitCodes.OK)
        self.assertFalse(self.slave.wasCalled('buildFail'))
        self.assertEqual(
            [((changes_path,), {})], self.slave.addWaitingFile.calls)

        # Control returns to the DebianBuildManager in the UMOUNT state.
        self.assertUnmountsSanely()
        self.assertFalse(self.slave.wasCalled('buildFail'))

    def test_abort_sbuild(self):
        # Aborting sbuild kills processes in the chroot.
        self.startBuild()

        # Send an abort command.  The build manager reaps processes.
        self.buildmanager.abort()
        self.assertState(
            BinaryPackageBuildState.SBUILD,
            ['sharepath/slavebin/scan-for-processes', 'scan-for-processes',
             self.buildid], False)
        self.assertFalse(self.slave.wasCalled('buildFail'))

        # If reaping completes successfully, the build manager returns
        # control to the DebianBuildManager in the UMOUNT state.
        self.assertUnmountsSanely()
        self.assertFalse(self.slave.wasCalled('buildFail'))

    def test_abort_sbuild_fail(self):
        # If killing processes in the chroot hangs, the build manager does
        # its best to clean up and fails the builder.
        self.startBuild()
        sbuild_subprocess = self.buildmanager._subprocess

        # Send an abort command.  The build manager reaps processes.
        self.buildmanager.abort()
        self.assertState(
            BinaryPackageBuildState.SBUILD,
            ['sharepath/slavebin/scan-for-processes', 'scan-for-processes',
             self.buildid], False)
        self.assertFalse(self.slave.wasCalled('builderFail'))
        reap_subprocess = self.buildmanager._subprocess

        # If reaping fails, the builder is failed, sbuild is killed, and the
        # reaper is disconnected.
        self.clock.advance(120)
        self.assertTrue(self.slave.wasCalled('builderFail'))
        self.assertEqual(
            [(('KILL',), {})], sbuild_subprocess.transport.signalProcess.calls)
        self.assertNotEqual(
            [], sbuild_subprocess.transport.loseConnection.calls)
        self.assertNotEqual([], reap_subprocess.transport.loseConnection.calls)

        write_file(
            os.path.join(self.buildmanager._cachepath, 'buildlog'),
            "I am a build log.")

        # When sbuild exits, it does not reap processes again, but proceeds
        # directly to UMOUNT.
        self.buildmanager.iterate(128 + 9)  # SIGKILL
        self.assertState(
            BinaryPackageBuildState.UMOUNT,
            ['sharepath/slavebin/umount-chroot', 'umount-chroot',
             self.buildid], True)

    def test_abort_between_subprocesses(self):
        # If a build is aborted between subprocesses, the build manager
        # pretends that it was terminated by a signal.
        self.buildmanager.initiate(
            {'foo_1.dsc': ''}, 'chroot.tar.gz',
            {'distribution': 'ubuntu', 'suite': 'warty',
             'ogrecomponent': 'main'})

        self.buildmanager.abort()
        self.assertState(
            BinaryPackageBuildState.INIT,
            ['sharepath/slavebin/scan-for-processes', 'scan-for-processes',
             self.buildid], False)

        self.buildmanager.iterate(0)
        self.assertState(
            BinaryPackageBuildState.CLEANUP,
            ['sharepath/slavebin/remove-build', 'remove-build', self.buildid],
            True)
        self.assertFalse(self.slave.wasCalled('builderFail'))

    def test_missing_changes(self):
        # The build manager recovers if the expected .changes file does not
        # exist, and considers it a package build failure.
        self.startBuild()
        write_file(
            os.path.join(self.buildmanager._cachepath, 'buildlog'),
            "I am a build log.")
        build_dir = os.path.join(
            self.buildmanager.home, 'build-%s' % self.buildid)
        write_file(
            os.path.join(build_dir, 'foo_2_i386.changes'),
            "I am a changes file.")

        # After building the package, reap processes.
        self.assertScansSanely(SBuildExitCodes.OK)
        self.assertTrue(self.slave.wasCalled('buildFail'))
        self.assertEqual(
            [((os.path.join(build_dir, 'foo_1_i386.changes'),), {})],
            self.slave.addWaitingFile.calls)

        # Control returns to the DebianBuildManager in the UMOUNT state.
        self.assertUnmountsSanely()
        self.assertTrue(self.slave.wasCalled('buildFail'))

    def test_getAvailablePackages(self):
        # getAvailablePackages scans the correct set of files and returns
        # reasonable version information.
        apt_lists = os.path.join(self.chrootdir, "var", "lib", "apt", "lists")
        os.makedirs(apt_lists)
        write_file(
            os.path.join(
                apt_lists,
                "archive.ubuntu.com_ubuntu_trusty_main_binary-amd64_Packages"),
            dedent("""\
                Package: foo
                Version: 1.0
                Provides: virt

                Package: bar
                Version: 2.0
                Provides: versioned-virt (= 3.0)
                """))
        write_file(
            os.path.join(
                apt_lists,
                "archive.ubuntu.com_ubuntu_trusty-proposed_main_binary-amd64_"
                "Packages"),
            dedent("""\
                Package: foo
                Version: 1.1
                """))
        write_file(os.path.join(apt_lists, "other"), "some other stuff")
        expected = {
            "foo": set(["1.0", "1.1"]),
            "bar": set(["2.0"]),
            "virt": set([None]),
            "versioned-virt": set(["3.0"]),
            }
        self.assertEqual(expected, self.buildmanager.getAvailablePackages())

    def test_getBuildDepends_arch_dep(self):
        # getBuildDepends returns only Build-Depends for
        # architecture-dependent builds.
        dscpath = os.path.join(
            self.working_dir, "build-%s" % self.buildid, "foo.dsc")
        write_file(
            dscpath,
            dedent("""\
                Package: foo
                Build-Depends: debhelper (>= 9~), bar | baz
                Build-Depends-Indep: texlive-base
                """))
        self.assertThat(
            self.buildmanager.getBuildDepends(dscpath, False),
            MatchesListwise([
                MatchesListwise([
                    ContainsDict({
                        "name": Equals("debhelper"),
                        "version": Equals((">=", "9~")),
                        }),
                    ]),
                MatchesListwise([
                    ContainsDict({"name": Equals("bar"), "version": Is(None)}),
                    ContainsDict({"name": Equals("baz"), "version": Is(None)}),
                    ])]))

    def test_getBuildDepends_arch_indep(self):
        # getBuildDepends returns both Build-Depends and Build-Depends-Indep
        # for architecture-independent builds.
        dscpath = os.path.join(
            self.working_dir, "build-%s" % self.buildid, "foo.dsc")
        write_file(
            dscpath,
            dedent("""\
                Package: foo
                Build-Depends: debhelper (>= 9~), bar | baz
                Build-Depends-Indep: texlive-base
                """))
        self.assertThat(
            self.buildmanager.getBuildDepends(dscpath, True),
            MatchesListwise([
                MatchesListwise([
                    ContainsDict({
                        "name": Equals("debhelper"),
                        "version": Equals((">=", "9~")),
                        }),
                    ]),
                MatchesListwise([
                    ContainsDict({"name": Equals("bar"), "version": Is(None)}),
                    ContainsDict({"name": Equals("baz"), "version": Is(None)}),
                    ]),
                MatchesListwise([
                    ContainsDict({
                        "name": Equals("texlive-base"),
                        "version": Is(None),
                        }),
                    ])]))

    def test_getBuildDepends_missing_fields(self):
        # getBuildDepends tolerates missing fields.
        dscpath = os.path.join(
            self.working_dir, "build-%s" % self.buildid, "foo.dsc")
        write_file(dscpath, "Package: foo\n")
        self.assertEqual([], self.buildmanager.getBuildDepends(dscpath, True))

    def test_relationMatches_missing_package(self):
        # relationMatches returns False if a dependency's package name is
        # entirely missing.
        self.assertFalse(self.buildmanager.relationMatches(
            {"name": "foo", "version": (">=", "1")}, {"bar": set(["2"])}))

    def test_relationMatches_unversioned(self):
        # relationMatches returns True if a dependency's package name is
        # present and the dependency is unversioned.
        self.assertTrue(self.buildmanager.relationMatches(
            {"name": "foo", "version": None}, {"foo": set(["1"])}))

    def test_relationMatches_versioned(self):
        # relationMatches handles versioned dependencies correctly.
        for version, expected in (
            (("<<", "1"), False), (("<<", "1.1"), True),
            (("<=", "0.9"), False), (("<=", "1"), True),
            (("=", "1"), True), (("=", "2"), False),
            ((">=", "1"), True), ((">=", "1.1"), False),
            ((">>", "0.9"), True), ((">>", "1"), False),
            ):
            assert_method = self.assertTrue if expected else self.assertFalse
            assert_method(self.buildmanager.relationMatches(
                {"name": "foo", "version": version}, {"foo": set(["1"])}),
                "%s %s 1 was not %s" % (version[1], version[0], expected))

    def test_relationMatches_multiple_versions(self):
        # If multiple versions of a package are present, relationMatches
        # returns True for dependencies that match any of them.
        for version, expected in (
            (("=", "1"), True),
            (("=", "1.1"), True),
            (("=", "2"), False),
            ):
            assert_method = self.assertTrue if expected else self.assertFalse
            assert_method(self.buildmanager.relationMatches(
                {"name": "foo", "version": version},
                {"foo": set(["1", "1.1"])}))

    def test_relationMatches_unversioned_virtual(self):
        # Unversioned dependencies match an unversioned virtual package, but
        # versioned dependencies do not.
        for version, expected in ((None, True), ((">=", "1"), False)):
            assert_method = self.assertTrue if expected else self.assertFalse
            assert_method(self.buildmanager.relationMatches(
                {"name": "foo", "version": version},
                {"foo": set([None])}))

    def test_analyseDepWait_all_satisfied(self):
        # If all direct build-dependencies are satisfied, analyseDepWait
        # returns None.
        self.assertIsNone(self.buildmanager.analyseDepWait(
            PkgRelation.parse_relations("debhelper, foo (>= 1)"),
            {"debhelper": set(["9"]), "foo": set(["1"])}))

    def test_analyseDepWait_unsatisfied(self):
        # If some direct build-dependencies are unsatisfied, analyseDepWait
        # returns a stringified representation of them.
        self.assertEqual(
            "foo (>= 1), bar (<< 1) | bar (>= 2)",
            self.buildmanager.analyseDepWait(
                PkgRelation.parse_relations(
                    "debhelper (>= 9~), foo (>= 1), bar (<< 1) | bar (>= 2)"),
                {"debhelper": set(["9"]), "bar": set(["1", "1.5"])}))

    def test_analyseDepWait_strips_arch_restrictions(self):
        # analyseDepWait removes architecture restrictions (e.g. "[amd64]")
        # from the unsatisfied build-dependencies it returns, and only
        # returns those relevant to the current architecture.
        self.buildmanager.initiate(
            {'foo_1.dsc': ''}, 'chroot.tar.gz',
            {'distribution': 'ubuntu', 'suite': 'warty',
             'ogrecomponent': 'main', 'arch_tag': 'i386'})
        self.assertEqual(
            "foo (>= 1)",
            self.buildmanager.analyseDepWait(
                PkgRelation.parse_relations(
                    "foo (>= 1) [any-i386], bar (>= 1) [amd64]"),
                {}))

    def test_analyseDepWait_strips_arch_qualifications(self):
        # analyseDepWait removes architecture qualifications (e.g. ":any")
        # from the unsatisfied build-dependencies it returns.
        self.buildmanager.initiate(
            {'foo_1.dsc': ''}, 'chroot.tar.gz',
            {'distribution': 'ubuntu', 'suite': 'warty',
             'ogrecomponent': 'main', 'arch_tag': 'i386'})
        self.assertEqual(
            "foo",
            self.buildmanager.analyseDepWait(
                PkgRelation.parse_relations("foo:any, bar:any"),
                {"bar": set(["1"])}))

    def test_analyseDepWait_strips_restrictions(self):
        # analyseDepWait removes restrictions (e.g. "<stage1>") from the
        # unsatisfied build-dependencies it returns, and only returns those
        # that evaluate to true when no build profiles are active.
        self.buildmanager.initiate(
            {'foo_1.dsc': ''}, 'chroot.tar.gz',
            {'distribution': 'ubuntu', 'suite': 'warty',
             'ogrecomponent': 'main', 'arch_tag': 'i386'})
        self.assertEqual(
            "foo",
            self.buildmanager.analyseDepWait(
                PkgRelation.parse_relations("foo <!nocheck>, bar <stage1>"),
                {}))

    def startDepFail(self, error, dscname=''):
        self.startBuild(dscname=dscname)
        write_file(
            os.path.join(self.buildmanager._cachepath, 'buildlog'),
            "The following packages have unmet dependencies:\n"
            + (" sbuild-build-depends-hello-dummy : Depends: %s\n" % error)
            + "E: Unable to correct problems, you have held broken packages.\n"
            + ("a" * 4096) + "\n"
            + "Fail-Stage: install-deps\n")

    def assertMatchesDepfail(self, error, dep):
        self.startDepFail(error)
        self.assertScansSanely(SBuildExitCodes.GIVENBACK)
        self.assertUnmountsSanely()
        if dep is not None:
            self.assertFalse(self.slave.wasCalled('buildFail'))
            self.assertEqual([((dep,), {})], self.slave.depFail.calls)
        else:
            self.assertFalse(self.slave.wasCalled('depFail'))
            self.assertTrue(self.slave.wasCalled('buildFail'))

    def test_detects_depfail(self):
        # The build manager detects dependency installation failures.
        self.assertMatchesDepfail(
            "enoent but it is not installable", "enoent")

    def test_detects_versioned_depfail(self):
        # The build manager detects dependency installation failures.
        self.assertMatchesDepfail(
            "ebadver (< 2.0) but 3.0 is to be installed", "ebadver (< 2.0)")

    def test_detects_versioned_current_depfail(self):
        # The build manager detects dependency installation failures.
        self.assertMatchesDepfail(
            "ebadver (< 2.0) but 3.0 is installed", "ebadver (< 2.0)")

    def test_strips_depfail(self):
        # The build manager strips qualifications and restrictions from
        # dependency installation failures.
        self.assertMatchesDepfail(
            "ebadver:any (>= 3.0) but 2.0 is installed", "ebadver (>= 3.0)")

    def test_uninstallable_deps_analysis_failure(self):
        # If there are uninstallable build-dependencies and analysis can't
        # find any missing direct build-dependencies, the build manager
        # fails the build as it doesn't have a condition on which it can
        # automatically retry later.
        write_file(
            os.path.join(self.buildmanager._cachepath, "123"),
            dedent("""\
                Package: foo
                Version: 1
                Build-Depends: uninstallable (>= 1)
                """))
        self.startDepFail(
            "uninstallable (>= 1) but it is not going to be installed",
            dscname="123")
        apt_lists = os.path.join(self.chrootdir, "var", "lib", "apt", "lists")
        os.makedirs(apt_lists)
        write_file(
            os.path.join(apt_lists, "archive_Packages"),
            dedent("""\
                Package: uninstallable
                Version: 1
                """))
        self.assertScansSanely(SBuildExitCodes.GIVENBACK)
        self.assertUnmountsSanely()
        self.assertFalse(self.slave.wasCalled('depFail'))
        self.assertTrue(self.slave.wasCalled('buildFail'))

    def test_uninstallable_deps_analysis_depfail(self):
        # If there are uninstallable build-dependencies and analysis reports
        # some missing direct build-dependencies, the build manager marks
        # the build as DEPFAIL.
        write_file(
            os.path.join(self.buildmanager._cachepath, "123"),
            dedent("""\
                Package: foo
                Version: 1
                Build-Depends: ebadver (>= 2)
                """))
        self.startDepFail(
            "ebadver (>= 2) but it is not going to be installed",
            dscname="123")
        apt_lists = os.path.join(self.chrootdir, "var", "lib", "apt", "lists")
        os.makedirs(apt_lists)
        write_file(
            os.path.join(apt_lists, "archive_Packages"),
            dedent("""\
                Package: ebadver
                Version: 1
                """))
        self.assertScansSanely(SBuildExitCodes.GIVENBACK)
        self.assertUnmountsSanely()
        self.assertFalse(self.slave.wasCalled('buildFail'))
        self.assertEqual([(("ebadver (>= 2)",), {})], self.slave.depFail.calls)

    def test_uninstallable_deps_analysis_mixed_depfail(self):
        # If there is a mix of definite and dubious dep-wait output, then
        # the build manager analyses the situation rather than trusting just
        # the definite information.
        write_file(
            os.path.join(self.buildmanager._cachepath, "123"),
            dedent("""\
                Package: foo
                Version: 1
                Build-Depends: ebadver (>= 2), uninstallable
                """))
        self.startBuild(dscname="123")
        write_file(
            os.path.join(self.buildmanager._cachepath, 'buildlog'),
            "The following packages have unmet dependencies:\n"
            + (" sbuild-build-depends-hello-dummy : Depends: ebadver (>= 2) "
               "but it is not going to be installed\n")
            + ("                                    Depends: uninstallable "
               "but it is not installable\n")
            + "E: Unable to correct problems, you have held broken packages.\n"
            + ("a" * 4096) + "\n"
            + "Fail-Stage: install-deps\n")
        apt_lists = os.path.join(self.chrootdir, "var", "lib", "apt", "lists")
        os.makedirs(apt_lists)
        write_file(
            os.path.join(apt_lists, "archive_Packages"),
            dedent("""\
                Package: ebadver
                Version: 1

                Package: uninstallable
                Version: 1
                """))
        self.assertScansSanely(SBuildExitCodes.GIVENBACK)
        self.assertUnmountsSanely()
        self.assertFalse(self.slave.wasCalled('buildFail'))
        self.assertEqual([(("ebadver (>= 2)",), {})], self.slave.depFail.calls)

    def test_depfail_with_unknown_error_converted_to_packagefail(self):
        # The build manager converts a DEPFAIL to a PACKAGEFAIL if the
        # missing dependency can't be determined from the log.
        self.startBuild()
        write_file(
            os.path.join(self.buildmanager._cachepath, 'buildlog'),
            "E: Everything is broken.\n")

        self.assertScansSanely(SBuildExitCodes.GIVENBACK)
        self.assertTrue(self.slave.wasCalled('buildFail'))
        self.assertFalse(self.slave.wasCalled('depFail'))



# Copyright 2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import re


non_meta_re = re.compile(r'^[a-zA-Z0-9+,./:=@_-]+$')

def shell_escape(arg):
    if non_meta_re.match(arg):
        return arg
    else:
        return "'%s'" % arg.replace("'", "'\\''")


linux32_arches = [
    "armel",
    "armhf",
    "hppa",
    "i386",
    "lpia",
    "mips",
    "mipsel",
    "powerpc",
    "s390",
    "sparc",
    ]
linux64_arches = [
    "alpha",
    "amd64",
    "arm64",
    "hppa64",
    "ia64",
    "ppc64",
    "ppc64el",
    "s390x",
    "sparc64",
    "x32",
    ]


def set_personality(arch, args):
    if arch in linux32_arches:
        return ["linux32"] + args
    elif arch in linux64_arches:
        return ["linux64"] + args
    else:
        return args



# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import doctest

def test_suite():
    return doctest.DocTestSuite('lpbuildd.tests.harness')




# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
# pylint: disable-msg=E1002

"""The manager class for building packages from recipes."""

import os
import re

from lpbuildd.debian import (
    DebianBuildManager,
    DebianBuildState,
    get_build_path,
)
RETCODE_SUCCESS = 0
RETCODE_FAILURE_INSTALL = 200
RETCODE_FAILURE_BUILD_TREE = 201
RETCODE_FAILURE_INSTALL_BUILD_DEPS = 202
RETCODE_FAILURE_BUILD_SOURCE_PACKAGE = 203


def splat_file(path, contents):
    """Write a string to the specified path.

    :param path: The path to store the string in.
    :param contents: The string to write to the file.
    """
    file_obj = open(path, 'w')
    try:
        file_obj.write(contents)
    finally:
        file_obj.close()


def get_chroot_path(home, build_id, *extra):
    """Return a path within the chroot.

    :param home: The user's home directory.
    :param build_id: The build_id of the build.
    :param extra: Additional path elements.
    """
    return get_build_path(
        home, build_id, 'chroot-autobuild', os.environ['HOME'][1:], *extra)


class SourcePackageRecipeBuildState(DebianBuildState):
    """The set of states that a recipe build can be in."""
    BUILD_RECIPE = "BUILD_RECIPE"


class SourcePackageRecipeBuildManager(DebianBuildManager):
    """Build a source package from a bzr-builder recipe."""

    initial_build_state = SourcePackageRecipeBuildState.BUILD_RECIPE

    def __init__(self, slave, buildid):
        """Constructor.

        :param slave: A build slave device.
        :param buildid: The id of the build (a str).
        """
        DebianBuildManager.__init__(self, slave, buildid)
        self.build_recipe_path = os.path.join(self._slavebin, "buildrecipe")

    def initiate(self, files, chroot, extra_args):
        """Initiate a build with a given set of files and chroot.

        :param files: The files sent by the manager with the request.
        :param chroot: The sha1sum of the chroot to use.
        :param extra_args: A dict of extra arguments.
        """
        self.recipe_text = extra_args['recipe_text']
        self.suite = extra_args['suite']
        self.component = extra_args['ogrecomponent']
        self.author_name = extra_args['author_name']
        self.author_email = extra_args['author_email']
        self.archive_purpose = extra_args['archive_purpose']
        self.distroseries_name = extra_args['distroseries_name']

        super(SourcePackageRecipeBuildManager, self).initiate(
            files, chroot, extra_args)

    def doRunBuild(self):
        """Run the build process to build the source package."""
        os.makedirs(get_chroot_path(self.home, self._buildid, 'work'))
        recipe_path = get_chroot_path(self.home, self._buildid, 'work/recipe')
        splat_file(recipe_path, self.recipe_text)
        args = [
            "buildrecipe", self._buildid, self.author_name.encode('utf-8'),
            self.author_email, self.suite, self.distroseries_name,
            self.component, self.archive_purpose]
        self.runSubProcess(self.build_recipe_path, args)

    def iterate_BUILD_RECIPE(self, retcode):
        """Move from BUILD_RECIPE to the next logical state."""
        if retcode == RETCODE_SUCCESS:
            self.gatherResults()
            print("Returning build status: OK")
        elif retcode == RETCODE_FAILURE_INSTALL_BUILD_DEPS:
            if not self.alreadyfailed:
                rx = (
                    'The following packages have unmet dependencies:\n'
                    '.*: Depends: ([^ ]*( \([^)]*\))?)')
                _, mo = self.searchLogContents([[rx, re.M]])
                if mo:
                    self._slave.depFail(mo.group(1))
                    print("Returning build status: DEPFAIL")
                    print("Dependencies: " + mo.group(1))
                else:
                    print("Returning build status: Build failed")
                    self._slave.buildFail()
            self.alreadyfailed = True
        elif (
            retcode >= RETCODE_FAILURE_INSTALL and
            retcode <= RETCODE_FAILURE_BUILD_SOURCE_PACKAGE):
            # XXX AaronBentley 2009-01-13: We should handle depwait separately
            if not self.alreadyfailed:
                self._slave.buildFail()
                print("Returning build status: Build failed.")
            self.alreadyfailed = True
        else:
            if not self.alreadyfailed:
                self._slave.builderFail()
                print("Returning build status: Builder failed.")
            self.alreadyfailed = True
        self.doReapProcesses(self._state)

    def iterateReap_BUILD_RECIPE(self, retcode):
        """Finished reaping after recipe building."""
        self._state = DebianBuildState.UMOUNT
        self.doUnmounting()

    def getChangesFilename(self):
        """Return the path to the changes file."""
        work_path = get_build_path(self.home, self._buildid)
        for name in os.listdir(work_path):
            if name.endswith('_source.changes'):
                return os.path.join(work_path, name)

    def gatherResults(self):
        """Gather the results of the build and add them to the file cache.

        The primary file we care about is the .changes file.
        The manifest is also a useful record.
        """
        DebianBuildManager.gatherResults(self)
        self._slave.addWaitingFile(get_build_path(
            self.home, self._buildid, 'manifest'))



# Copyright 2009-2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

import signal
import sys

from twisted.internet import reactor
from twisted.python import (
    log,
    logfile,
    )
from zope.interface import implements


class RotatableFileLogObserver:
    """A log observer that uses a log file and reopens it on SIGHUP."""

    implements(log.ILogObserver)

    def __init__(self, logfilepath):
        """Set up the logfile and possible signal handler.

        Installs the signal handler for SIGHUP to make the process re-open
        the log file.

        :param logfilepath: The path to the logfile. If None, stdout is used
            for logging and no signal handler will be installed.
        """
        if logfilepath is None:
            logFile = sys.stdout
        else:
            logFile = logfile.LogFile.fromFullPath(
                logfilepath, rotateLength=None)
            # Override if signal is set to None or SIG_DFL (0)
            if not signal.getsignal(signal.SIGHUP):
                def signalHandler(signal, frame):
                    reactor.callFromThread(logFile.reopen)
                signal.signal(signal.SIGHUP, signalHandler)
        self.observer = log.FileLogObserver(logFile)

    def __call__(self, eventDict):
        self.observer.emit(eventDict)



# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import os
import shutil
import tempfile
from textwrap import dedent

from testtools import TestCase

from lpbuildd.tests.fakeslave import FakeSlave
from lpbuildd.sourcepackagerecipe import (
    RETCODE_FAILURE_INSTALL_BUILD_DEPS,
    SourcePackageRecipeBuildManager,
    SourcePackageRecipeBuildState,
    )


class MockBuildManager(SourcePackageRecipeBuildManager):
    def __init__(self, *args, **kwargs):
        super(MockBuildManager, self).__init__(*args, **kwargs)
        self.commands = []
        self.iterators = []

    def runSubProcess(self, path, command, iterate=None):
        self.commands.append([path]+command)
        if iterate is None:
            iterate = self.iterate
        self.iterators.append(iterate)
        return 0


class TestSourcePackageRecipeBuildManagerIteration(TestCase):
    """Run SourcePackageRecipeBuildManager through its iteration steps."""
    def setUp(self):
        super(TestSourcePackageRecipeBuildManagerIteration, self).setUp()
        self.working_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.working_dir))
        slave_dir = os.path.join(self.working_dir, 'slave')
        home_dir = os.path.join(self.working_dir, 'home')
        for dir in (slave_dir, home_dir):
            os.mkdir(dir)
        self.slave = FakeSlave(slave_dir)
        self.buildid = '123'
        self.buildmanager = MockBuildManager(self.slave, self.buildid)
        self.buildmanager.home = home_dir
        self.buildmanager._cachepath = self.slave._cachepath
        self.chrootdir = os.path.join(
            home_dir, 'build-%s' % self.buildid, 'chroot-autobuild')

    def getState(self):
        """Retrieve build manager's state."""
        return self.buildmanager._state

    def startBuild(self):
        # The build manager's iterate() kicks off the consecutive states
        # after INIT.
        extra_args = {
            'recipe_text': dedent("""\
                # bzr-builder format 0.2 deb-version {debupstream}-0~{revno}
                http://bazaar.launchpad.dev/~ppa-user/+junk/wakeonlan"""),
            'suite': 'maverick',
            'ogrecomponent': 'universe',
            'author_name': 'Steve\u1234',
            'author_email': 'stevea@example.org',
            'archive_purpose': 'puppies',
            'distroseries_name': 'maverick',
            'archives': [
                'deb http://archive.ubuntu.com/ubuntu maverick main universe',
                'deb http://ppa.launchpad.net/launchpad/bzr-builder-dev/'
                    'ubuntu main',
                ],
            }
        self.buildmanager.initiate({}, 'chroot.tar.gz', extra_args)

        # Skip states that are done in DebianBuildManager to the state
        # directly before BUILD_RECIPE.
        self.buildmanager._state = SourcePackageRecipeBuildState.UPDATE

        # BUILD_RECIPE: Run the slave's payload to build the source package.
        self.buildmanager.iterate(0)
        self.assertEqual(
            SourcePackageRecipeBuildState.BUILD_RECIPE, self.getState())
        expected_command = [
            'sharepath/slavebin/buildrecipe', 'buildrecipe', self.buildid,
            'Steve\u1234'.encode('utf-8'), 'stevea@example.org',
            'maverick', 'maverick', 'universe', 'puppies',
            ]
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('chrootFail'))

    def test_iterate(self):
        # The build manager iterates a normal build from start to finish.
        self.startBuild()

        log_path = os.path.join(self.buildmanager._cachepath, 'buildlog')
        log = open(log_path, 'w')
        log.write("I am a build log.")
        log.close()

        changes_path = os.path.join(
            self.buildmanager.home, 'build-%s' % self.buildid,
            'foo_1_source.changes')
        changes = open(changes_path, 'w')
        changes.write("I am a changes file.")
        changes.close()

        manifest_path = os.path.join(
            self.buildmanager.home, 'build-%s' % self.buildid, 'manifest')
        manifest = open(manifest_path, 'w')
        manifest.write("I am a manifest file.")
        manifest.close()

        # After building the package, reap processes.
        self.buildmanager.iterate(0)
        expected_command = [
            'sharepath/slavebin/scan-for-processes', 'scan-for-processes',
            self.buildid,
            ]
        self.assertEqual(
            SourcePackageRecipeBuildState.BUILD_RECIPE, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertNotEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('buildFail'))
        self.assertEqual(
            [((changes_path,), {}), ((manifest_path,), {})],
            self.slave.addWaitingFile.calls)

        # Control returns to the DebianBuildManager in the UMOUNT state.
        self.buildmanager.iterateReap(self.getState(), 0)
        expected_command = [
            'sharepath/slavebin/umount-chroot', 'umount-chroot', self.buildid
            ]
        self.assertEqual(SourcePackageRecipeBuildState.UMOUNT, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('buildFail'))

    def test_iterate_BUILD_RECIPE_install_build_deps_depfail(self):
        # The build manager can detect dependency wait states.
        self.startBuild()

        log_path = os.path.join(self.buildmanager._cachepath, 'buildlog')
        log = open(log_path, 'w')
        log.write(
            "The following packages have unmet dependencies:\n"
            " pbuilder-satisfydepends-dummy : Depends: base-files (>= 1000)"
            " but it is not going to be installed.\n")
        log.close()

        # The buildmanager calls depFail correctly and reaps processes.
        self.buildmanager.iterate(RETCODE_FAILURE_INSTALL_BUILD_DEPS)
        expected_command = [
            'sharepath/slavebin/scan-for-processes', 'scan-for-processes',
            self.buildid,
            ]
        self.assertEqual(
            SourcePackageRecipeBuildState.BUILD_RECIPE, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertNotEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('buildFail'))
        self.assertEqual(
            [(("base-files (>= 1000)",), {})], self.slave.depFail.calls)

        # Control returns to the DebianBuildManager in the UMOUNT state.
        self.buildmanager.iterateReap(self.getState(), 0)
        expected_command = [
            'sharepath/slavebin/umount-chroot', 'umount-chroot', self.buildid,
            ]
        self.assertEqual(SourcePackageRecipeBuildState.UMOUNT, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertFalse(self.slave.wasCalled('buildFail'))

    def test_iterate_BUILD_RECIPE_install_build_deps_buildfail(self):
        # If the build manager cannot detect a dependency wait from a
        # build-dependency installation failure, it fails the build.
        self.startBuild()

        log_path = os.path.join(self.buildmanager._cachepath, 'buildlog')
        log = open(log_path, 'w')
        log.write("I am a failing build log.")
        log.close()

        # The buildmanager calls buildFail correctly and reaps processes.
        self.buildmanager.iterate(RETCODE_FAILURE_INSTALL_BUILD_DEPS)
        expected_command = [
            'sharepath/slavebin/scan-for-processes', 'scan-for-processes',
            self.buildid,
            ]
        self.assertEqual(
            SourcePackageRecipeBuildState.BUILD_RECIPE, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertNotEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])
        self.assertTrue(self.slave.wasCalled('buildFail'))
        self.assertFalse(self.slave.wasCalled('depFail'))

        # Control returns to the DebianBuildManager in the UMOUNT state.
        self.buildmanager.iterateReap(self.getState(), 0)
        expected_command = [
            'sharepath/slavebin/umount-chroot', 'umount-chroot', self.buildid,
            ]
        self.assertEqual(SourcePackageRecipeBuildState.UMOUNT, self.getState())
        self.assertEqual(expected_command, self.buildmanager.commands[-1])
        self.assertEqual(
            self.buildmanager.iterate, self.buildmanager.iterators[-1])



# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from testtools import TestCase

from lpbuildd.check_implicit_pointer_functions import implicit_pattern
from lpbuildd.check_implicit_pointer_functions import pointer_pattern


class TestPointerCheckRegexes(TestCase):

    def test_catches_pointer_from_integer_without_column_number(self):
        # Regex should match compiler errors that don't include the
        # column number.
        line = (
            "/build/buildd/gtk+3.0-3.0.0/./gtk/ubuntumenuproxymodule.c:94: "
            "warning: assignment makes pointer from integer without a cast")
        self.assertIsNot(None, pointer_pattern.match(line))

    def test_catches_pointer_from_integer_with_column_number(self):
        # Regex should match compiler errors that do include the
        # column number.
        line = (
            "/build/buildd/gtk+3.0-3.0.0/./gtk/ubuntumenuproxymodule.c:94:7: "
            "warning: assignment makes pointer from integer without a cast")
        self.assertIsNot(None, pointer_pattern.match(line))

    def test_catches_implicit_function_without_column_number(self):
        # Regex should match compiler errors that do include the
        # column number.
        line = (
            "/build/buildd/gtk+3.0-3.0.0/./gtk/ubuntumenuproxymodule.c:94: "
            "warning: implicit declaration of function 'foo'")
        self.assertIsNot(None, implicit_pattern.match(line))

    def test_catches_implicit_function_with_column_number(self):
        # Regex should match compiler errors that do include the
        # column number.
        line = (
            "/build/buildd/gtk+3.0-3.0.0/./gtk/ubuntumenuproxymodule.c:94:7: "
            "warning: implicit declaration of function 'foo'")
        self.assertIsNot(None, implicit_pattern.match(line))




#!/usr/bin/env python

# Copyright 2015 Canonical Ltd.  All rights reserved.
#
# This file is part of launchpad-buildd.
#
# launchpad-buildd is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, version 3 of the License.
#
# launchpad-buildd is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with launchpad-buildd.  If not, see <http://www.gnu.org/licenses/>.

import re
from textwrap import dedent

from setuptools import (
    find_packages,
    setup,
    )


changelog_heading = re.compile(r'\w[-+0-9a-z.]* \(([^\(\) \t]+)\)')

with open('debian/changelog') as changelog:
    line = changelog.readline()
    match = changelog_heading.match(line)
    if match is None:
        raise ValueError(
            "Failed to parse first line of debian/changelog: '%s'" % line)
    version = match.group(1)


setup(
    name='launchpad-buildd',
    version=version,
    description='Launchpad buildd slave',
    long_description=dedent("""
        The Launchpad buildd slave libraries.  The PyPI version of this
        package will not produce a complete installation on its own, and is
        mostly useful for testing other pieces of software against
        launchpad-buildd; for a real Launchpad buildd slave, install the
        launchpad-buildd package from ppa:launchpad/ubuntu/ppa.
        """).strip(),
    url='https://launchpad.net/launchpad-buildd',
    packages=find_packages(),
    package_data={
        'lpbuildd': [
            'tests/buildd-slave-test.conf',
            ],
        },
    maintainer='Launchpad Developers',
    maintainer_email='launchpad-dev@lists.launchpad.net',
    license='Affero GPL v3',
    install_requires=[
        'bzr',
        # XXX cjwatson 2015-11-04: This does in fact require python-apt, but
        # that's normally shipped as a system package and specifying it here
        # causes problems for Launchpad's build system.
        #'python-apt',
        'python-debian',
        'Twisted',
        'zope.interface',
        ],
    data_files=[
        ('', ['buildd-slave.tac', 'template-buildd-slave.conf']),
        ],
    test_suite='lpbuildd.tests',
    tests_require=[
        'fixtures',
        'testtools',
        'txfixtures',
        ],
    )
